#ifndef __HDE_H__  // Include this file only once!
#define __HDE_H__

#include "../detri2/detri2.h"

namespace detri2
{

//==============================================================================

class VertexData 
{
 public:
  REAL v[4]; // f(x,y), f_x, f_y, -1.
  REAL uh0, uh1; // Previous solutions of wave equation

  void init() {
    v[0] = v[1] = v[2] = v[3] = 0.0;
    uh0 = uh1 = 0.0;
  }

  VertexData() {init();}

  void setSolution(REAL sol)             {v[0] = sol;}
  void setGradientX(REAL xComp)          {v[1] = xComp;}
  void setGradientY(REAL yComp)          {v[2] = yComp;}
  void setGradientZ(REAL zComp)          {v[3] = zComp;}

  REAL getSolution()                     {return(v[0]);}
  REAL getGradientX()                    {return(v[1]);}
  REAL getGradientY()                    {return(v[2]);}
  REAL getGradientZ()                    {return(v[3]);}

  /*
  // MetricAnIso
  void set_a11(REAL a11)  {v[0] = a11;}
  void set_a21(REAL a21)  {v[1] = a21;}
  void set_a22(REAL a22)  {v[2] = a22;}
  REAL get_a11()          {return v[0];}
  REAL get_a21()          {return v[1];}
  REAL get_a22()          {return v[2];}
  */
};

typedef REAL (scalarFunType)(const Vertex * pt);

//==============================================================================

class HDE_Triangulation : public Triangulation
{
 private:

  arraypool *VertexDataPool;
  arraypool *OMT_domain_VertexDataPool;

 public:

  // Geometry information.
  REAL      geo_domain_diameter;
  REAL      geo_average_length;

  // Metric options
  int       op_use_metric; // default 0 (use metric distance)
  int       op_use_HDE_metric; // default 0 (use HDE mretic distance)
  //REAL      op_mtr_coef; // default = 1.0 (metric)
  //REAL      op_mtr_anisomax; // default = 1e+6 (metric)
  //REAL      geo_hmin, geo_hmax; // for metric

  // HDE options
  REAL      op_s1, op_s2, op_s3; // s factors, default = 0.
  REAL      op_max_fun_abs; // default =0.0
  REAL      op_max_grd_mag; // default =0.0

  // Mesh size and mesh quality options
  //REAL      op_target_length; // default = 0.
  //REAL      op_target_min_angle;  // default = 0 degree
  //REAL      op_target_max_angle;  // default = 180 degree
  //REAL      op_target_cos_min_angle;
  //REAL      op_target_cos_max_angle;

  // Mesh adaptation options
  int       op_max_size_iter; // default = 5.
  int       op_max_quality_iter; // default = 3.
  int       op_use_intpoints; // default = 2.
  int       op_use_smoothing; // default = 1.
  int       op_use_repairing; // default = 1.
  int       op_report_quality_only; // default = 0.

  // Edge splitting-collapse options.
  int       op_split_iter; // default = 10
  int       op_edge_bisect_iter; // default = 2.
  REAL      op_edge_split_factor; // default = 1.45
  REAL      op_edge_collapse_factor; // default = 0.3
  REAL      op_max_enc_angle; // default max_enc_ang=175 degree.
  REAL      op_min_enc_angle; // default min_enc_ang=5 degree.

  // Vertex smoothing options
  int       op_smooth_iter; // default = 10.
  REAL      op_smooth_alpha; // default = 0.01.

  // Output options
  int       op_save_inter_meshes; // default = 0.

  // A user-provided function, its gradients, and its Hessian matrix.
  scalarFunType *Func, *GradX, *GradY;
  scalarFunType *FunXX, *FunXY, *FunYY; // Hessian matrix

  // The choice of test functions of dteri2HDE.
  int       op_test_function;  // default = 0 (No function is chosen)

  // I/O
  int  HDE_parse_commands(int argc, char* argv[]);
  double HDE_get_average_length(); // distmesh, to scale gradients
  //int  HDE_reconstruct_OMT_domain(REAL *sol, REAL *gradx, REAL *grady);
  int HDE_prepare();

  int  HDE_interpolate_node_data(Vertex *pt, TriEdge &E_dm,
                                 REAL* val, // mesh size
                                 REAL* sol, REAL* grdx, REAL* grdy,
                                 REAL* uh0, REAL* uh1);
  int  HDE_set_VertexData(Vertex *v);

  //int  HDE_metric_prepare();
  //int  HDE_prepare();

  // HDE length, angles.
  REAL HDE_dot(Vertex* v1, Vertex* v2);
  REAL HDE_dist(Vertex* v1, Vertex* v2);
  REAL HDE_angle(Vertex* v0, Vertex* v1, Vertex* v2); // angle at v0

  // Mesh adapt operations
  int  HDE_regular_test(Vertex* pa, Vertex* pb, Vertex* pc, Vertex* pd);
  int  HDE_lawson_flip(Vertex *pt, arraypool *fqueue);

  REAL HDE_get_vertex_size(Vertex *v);
  int  HDE_remove_point(Vertex *pt, arraypool* fqueue);
  int  HDE_coarsen_mesh();
  
  int  HDE_get_edge_bisector(Vertex *pa, Vertex *pb, Vertex *mid, REAL *lo, REAL *hi, int iter);
  int  HDE_split_edge(TriEdge &E, double t, Vertex **ppt, arraypool* fqueue);
  int  HDE_split_edges();

  int  HDE_move_interior_point(Vertex *pt);
  int  HDE_get_distmesh_point(Vertex *mesh_vertex, Vertex *movept);
  int  HDE_smooth_vertices();
  
  // Mesh repair operations
  REAL HDE_get_max_angle(TriEdge &E);
  REAL HDE_get_min_angle(TriEdge &E);
  int  HDE_check_triangle_inequality(TriEdge &E);
  int  HDE_check_triangle_quality(TriEdge &E);
  int  HDE_repair_large_angle(Triang *tri, int*, arraypool* fqueue);
  int  HDE_repair_small_angle(Triang *tri, int*, arraypool* fqueue);
  //int  HDE_repair_large_angles();
  //int  HDE_repair_small_angles();
  int  HDE_repair_triangles(); // avoid too streched triangles.

  // Mesh quality statistics
  REAL HDE_get_interpolation_errors(Triang* tri, REAL &L2err, REAL &H1err);
  REAL HDE_get_interpolation_error(REAL* L2err, REAL* H1err);
  void HDE_quality_statistic();

  void HDE_mesh_adapt();

  void writeEmbeddedMeshLifted(int meshidx);  // save function
  void writeEmbeddedEdgeLength(int meshidx);
  void writeEmbeddedTriangleMaxAngle(int meshidx);
  void writeEmbeddedTriangleMinAngle(int meshidx);
  void writeEmbeddedTriangleError(int meshidx);
  //void save_m();

  // Debug functions.
  void HDE_db_vrt(int v);
  REAL HDE_db_dist(int v1, int v2);
  REAL HDE_db_angle(int v1, int v2, int v3);
  void HDE_db_tri(int v1, int v2, int v3);
  void HDE_dump_splitting_edges(arraypool* edge_list, void* fout);
  void HDE_dump_triangles(arraypool* tri_list, void* fout);
  void HDE_save_metric(); // for op_use_metric

 public:

  void HDE_initialize();
  void HDE_clean();

  HDE_Triangulation();
  ~HDE_Triangulation();
};

}; // namespace detri2

#endif // #ifndef __HDE_H__
