
#include <QPainter>
#include <QKeyEvent>
#include <QFileDialog>
#include <QInputDialog>
#include <QMessageBox>
#include <QDebug>
#include <math.h>

#define MACOSX

#ifdef MACOSX
  #include "OpenGL/glu.h"
#else
  #include <gl/gl.h>
  #include <gl/glu.h>
#endif

#include "view3d.h"
#include "scribblearea.h"

using namespace detri2;

View3dWidget::View3dWidget(void* scribbleArea, QWidget *parent)
{
    QToolBar* toolbar_top = new QToolBar();
    addToolBar(Qt::TopToolBarArea, toolbar_top);
    QToolBar* toolbar_left = new QToolBar();
    addToolBar(Qt::LeftToolBarArea, toolbar_left);

    _glpanel = new OGLWidget(scribbleArea, 
                             menuBar(), 
                             statusBar(), 
                             toolbar_top, 
                             toolbar_left, 
                             this);
    setCentralWidget(_glpanel);

    setWindowTitle(tr("3d Viewer"));

    setMinimumSize(200, 200);
    resize(700, 500);
}

//==============================================================================
OGLWidget::OGLWidget(void* scribbleArea,
                     QMenuBar* menubar,
                     QStatusBar* statusbar,
                     QToolBar* toolbar_top,
                     QToolBar* toolbar_left,
                     QWidget *parent)
    : QOpenGLWidget(parent)
{
    _menubar = menubar;

    _statusbar = statusbar;
    m_statusLeft = new QLabel("Left", this);
    m_statusLeft->setFrameStyle(QFrame::Panel | QFrame::Sunken);
    m_statusMiddle = new QLabel("Middle", this);
    m_statusMiddle->setFrameStyle(QFrame::Panel | QFrame::Sunken);
    m_statusRight = new QLabel("Right", this);
    m_statusRight->setFrameStyle(QFrame::Panel | QFrame::Sunken);

    _statusbar->addPermanentWidget(m_statusLeft, 1);
    _statusbar->addPermanentWidget(m_statusMiddle, 2);
    _statusbar->addPermanentWidget(m_statusRight, 4);

    _toolbar_top = toolbar_top;
    _toolbar_left = toolbar_left;

    createActions();
    createMenus();

    // Set up a context menu opoped up by a mouse right-click.
    // http://www.setnode.com/blog/right-click-context-menus-with-qt/
    setContextMenuPolicy(Qt::CustomContextMenu);
    connect(this, SIGNAL(customContextMenuRequested(QPoint)),
            this, SLOT(ShowContextMenu(QPoint)));

    _disc = NULL;

    _zshift = 0.25;
    _ccent[0] = _ccent[1] = 0.0;
    _radius = 1.0;

    //_a11 = 2.0; // 1.0
    //_a21 = 0.0;
    //_a22 = 1.0;

    _ortho = true;
    _w = _h = 0;

    _transparency = 0.6;
    _shrink = 0.9;
    _plane_normal[0] = 0.;
    _plane_normal[1] = 0.;
    _plane_normal[2] = 1.;
    _plane_dir = 2;
    _plane_d = 0.5;

    _mouse_act = eRot;

    _scribbleArea = scribbleArea;
    //_mesh = NULL;

    // Selection test.
    _pos.init();
    _ray.init();
    _sel_vrts   = new arraypool(sizeof(Vertex), 8);
    _sel_edges  = new arraypool(sizeof(Vertex)*2, 8);

    _draw_object=NULL;  // A callback function (set in plot3d())
    _object_valid=0;
    _list_base=0;

    //_file_path = QDir::home().dirName(); // home directory
    //reset_volume();
    reset_view_transformations();
    reset_view_preferences();

    setFocus();  // get the keybordfocus for capturing keyPressEvent().
}

OGLWidget::~OGLWidget()
{
    if (_disc != NULL) delete _disc;
    delete _sel_vrts;
    delete _sel_edges;
}

//=================================================================================

// when bbox changed
void OGLWidget::reset_volume()
{
    //printf("debug:  =====reset_volume()===== is called.\n");
    ScribbleArea *sa =  (ScribbleArea *) _scribbleArea;
    //Triangulation *_mesh = ((ScribbleArea *) _scribbleArea)->_tria;
    Triangulation *_geom = sa->_geom;

    _min[0] = _geom->io_xmin;
    _min[1] = _geom->io_ymin;
    _max[0] = _geom->io_xmax;
    _max[1] = _geom->io_ymax;
    // Default.
    _min[2] = -1.; //zmin;
    _max[2] = 1.; // zmax;

    //double zmin = 1.e30, zmax=-1.e30;
    double zmin = 0, zmax = 0;

    triangulate_ellipse(); ////////////////////////////
    //triangulate_disc();

    double x = _disc->in_vrts[0].crd[0];
    double y = _disc->in_vrts[0].crd[1];
    //double z = x*x + y*y;
    double z = sa->_a11 * x*x + 2*sa->_a21*x*y + sa->_a22 * y*y;
    if (_max[2] < (z + _zshift)) {
      _max[2] = (z + _zshift);
    }

    qDebug() << "  xmin=" <<_min[0]<< " xmax="<<_max[0];
    qDebug() << "  ymin=" <<_min[1]<< " ymax="<<_max[1];
    qDebug() << "  zmin=" <<_min[2]<< " zmax="<<_max[2];

    _dmax = _max[0] - _min[0];
    if (_dmax < (_max[1] - _min[1])) {_dmax = (_min[1] - _min[1]);}
    if (_dmax < (zmax - zmin)) {_dmax = (zmax - zmin);}

    qDebug() << "  dmax=" << _dmax;
}

void  OGLWidget::reset_view_transformations()
{
    _rotX = _rotY = _rotZ = 0.0;   // roate angles, in degrees
    _tranX = _tranY = 0.0;
    _zoom = 1.0;
    _zoom_max = 10.0;
    _zNear = 4.; _zFar = 6.0;
    _prevX = _prevY = 0; // last mouse position in screen coordinates
}

void  OGLWidget::reset_view_preferences()
{
    if (_scribbleArea == NULL) {
      _vertex_size = 8;
      _edge_width = 1;
      _segment_width = 2;

      _bgcolor = Qt::black;
      _vertexcolor = Qt::gray;
      _edgecolor.setRgb(128, 128, 128);
      _facetcolor = Qt::cyan;
      _selectcolor = Qt::red;
      _circlecolor.setRgb(244, 66, 235);
      _weightcolor = Qt::red;
      _polecolor = Qt::blue;
      _polarcolor = Qt::yellow;
    } else {
      ScribbleArea *sa = (ScribbleArea *) _scribbleArea;
      _vertex_size = sa->_v3d_vertex_size;
      _edge_width = sa->_v3d_edge_width;
      _segment_width = sa->_v3d_segment_width;

      _bgcolor = sa->_v3d_bgcolor;
      _vertexcolor = sa->_v3d_vertexcolor;
      _edgecolor = sa->_v3d_edgecolor; //.setRgb(128, 128, 128);
      _facetcolor = sa->_v3d_facetcolor; // Qt::cyan;
      _selectcolor = sa->_v3d_selectcolor; // Qt::red;
      _circlecolor = sa->_v3d_circlecolor; // .setRgb(244, 66, 235);
      _weightcolor = sa->_v3d_weightcolor; //  Qt::red;
      _polecolor = sa->_v3d_polecolor; // Qt::blue;
      _polarcolor = sa->_v3d_polarcolor; // Qt::yellow;
    }

    // RGB Color picker
    // https://www.w3schools.com/colors/colors_picker.asp
}

//=================================================================================

void OGLWidget::createActions()
{
    openAct = new QAction(tr("&Open..."), this);
    openAct->setShortcuts(QKeySequence::Open);
    openAct->setStatusTip(tr("Open a file contains a geometry or a mesh"));
    connect(openAct, &QAction::triggered, this, &OGLWidget::openFile);

    saveAct = new QAction(tr("&Save As..."), this);
    saveAct->setShortcuts(QKeySequence::Save);
    saveAct->setStatusTip(tr("Save geometry/mesh/image to a file"));
    connect(saveAct, &QAction::triggered, this, &OGLWidget::saveFile);

    closeAct = new QAction(tr("Close"), this);
    closeAct->setShortcuts(QKeySequence::Close);
    connect(closeAct, &QAction::triggered, this, &OGLWidget::closeFile);

    // Mouse actions
    rotateAct = new QAction(tr("&Rotate"), this);
    rotateAct->setShortcut(QKeySequence(tr("R")));
    rotateAct->setStatusTip(tr("Rotate"));
    rotateAct->setCheckable(true);
    rotateAct->setChecked(true); // default
    connect(rotateAct, &QAction::triggered, this, &OGLWidget::mouse_act);

    transAct = new QAction(tr("&Move"), this);
    transAct->setShortcut(QKeySequence(tr("T")));
    transAct->setStatusTip(tr("Move"));
    transAct->setCheckable(true);
    connect(transAct, &QAction::triggered, this, &OGLWidget::mouse_act);

    shrinkAct = new QAction(tr("&Shrink"), this);
    shrinkAct->setShortcut(QKeySequence(tr("S")));
    shrinkAct->setStatusTip(tr("Shrink"));
    shrinkAct->setCheckable(true);
    connect(shrinkAct, &QAction::triggered, this, &OGLWidget::mouse_act);

    planeAct = new QAction(tr("&Plane"), this);
    planeAct->setShortcut(QKeySequence(tr("P")));
    planeAct->setStatusTip(tr("Plane"));
    planeAct->setCheckable(true);
    connect(planeAct, &QAction::triggered, this, &OGLWidget::mouse_act);

    selectAct = new QAction(tr("&Select"), this);
    //selectAct->setShortcut(QKeySequence(tr("S")));
    selectAct->setStatusTip(tr("Select"));
    selectAct->setCheckable(true);
    connect(selectAct, &QAction::triggered, this, &OGLWidget::mouse_act);

    // Mouse actions
    QActionGroup *actionGroup = new QActionGroup(this);
    actionGroup->addAction(rotateAct);
    actionGroup->addAction(transAct);
    //actionGroup->addAction(zoomRectAct);
    actionGroup->addAction(shrinkAct);
    actionGroup->addAction(planeAct);
    actionGroup->addAction(selectAct);

    // View options
    showVertices = new QAction(tr("Vertices"), this);
    //showVertices->setShortcut(QKeySequence(tr("Alt+P")));
    showVertices->setCheckable(true);
    showVertices->setChecked(true);
    connect(showVertices, &QAction::triggered, this, &OGLWidget::view_update_act);

    showVertexLabels = new QAction(tr("Vertex Labels"), this);
    //showVertexLabels->setShortcut(QKeySequence(tr("Alt+P")));
    showVertexLabels->setCheckable(true);
    showVertexLabels->setChecked(false);
    connect(showVertexLabels, &QAction::triggered, this, &OGLWidget::update_act);

    showEdges = new QAction(tr("Show edges"), this);
    //showEdges->setShortcut(QKeySequence(tr("Alt+E")));
    showEdges->setCheckable(true);
    showEdges->setChecked(true);
    connect(showEdges, &QAction::triggered, this, &OGLWidget::view_update_act);

    showSegments = new QAction(tr("Segments"), this);
    //showSegments->setShortcut(QKeySequence(tr("Alt+E")));
    showSegments->setCheckable(true);
    showSegments->setChecked(false);
    connect(showSegments, &QAction::triggered, this, &OGLWidget::view_update_act);

    showFacets = new QAction(tr("filled facets"), this);
    //showFacets->setShortcut(QKeySequence(tr("Alt+E")));
    showFacets->setCheckable(true);
    showFacets->setChecked(false); // default
    connect(showFacets, &QAction::triggered, this, &OGLWidget::view_update_act);

    showBbox = new QAction(tr("Bounding box"), this);
    //showBbox->setShortcut(QKeySequence(tr("Alt+E")));
    showBbox->setCheckable(true);
    showBbox->setChecked(true); // default
    connect(showBbox, &QAction::triggered, this, &OGLWidget::update_act);

    showPlane = new QAction(tr("Cut plane"), this);
    //showPlane->setShortcut(QKeySequence(tr("Alt+E")));
    showPlane->setCheckable(true);
    showPlane->setChecked(true); // default
    connect(showPlane, &QAction::triggered, this, &OGLWidget::update_act);

    showAxis = new QAction(tr("Axis"), this);
    //showAxis->setShortcut(QKeySequence(tr("Alt+E")));
    showAxis->setCheckable(true);
    showAxis->setChecked(true); // default
    connect(showAxis, &QAction::triggered, this, &OGLWidget::update_act);

    showTransparent = new QAction(tr("Transparent"), this);
    //showTransparent->setShortcut(QKeySequence(tr("Alt+E")));
    showTransparent->setCheckable(true);
    showTransparent->setChecked(false); // default
    connect(showTransparent, &QAction::triggered, this, &OGLWidget::view_update_act);

    enableShrink = new QAction(tr("Enable Shrink"), this);
    //enableShrink->setShortcut(QKeySequence(tr("Alt+E")));
    enableShrink->setCheckable(true);
    enableShrink->setChecked(true); // default
    connect(enableShrink, &QAction::triggered, this, &OGLWidget::view_update_act);

    enableCutPlane = new QAction(tr("Enable Cutplane"), this);
    //enableCutPlane->setShortcut(QKeySequence(tr("Alt+E")));
    enableCutPlane->setCheckable(true);
    enableCutPlane->setChecked(false); // default
    connect(enableCutPlane, &QAction::triggered, this, &OGLWidget::view_update_act);

    resetVolumeAct = new QAction(tr("Reset View Volume"), this);
    //resetVolumeAct->setShortcut(QKeySequence(tr("Alt+E")));
    //resetVolumeAct->setCheckable(true);
    //resetVolumeAct->setChecked(false); // default
    connect(resetVolumeAct, &QAction::triggered, this, &OGLWidget::reset_volume_act);

    invertCutPlane = new QAction(tr("Invert Cutplane"), this);
    //invertCutPlane->setShortcut(QKeySequence(tr("Alt+E")));
    invertCutPlane->setCheckable(true);
    invertCutPlane->setChecked(false); // default
    connect(invertCutPlane, &QAction::triggered, this, &OGLWidget::view_update_act);

    queryVertices = new QAction(tr("Query vertices"), this);
    queryVertices->setShortcut(QKeySequence(tr("Alt+V")));
    queryVertices->setCheckable(true);
    queryVertices->setChecked(true); // default
    connect(queryVertices, &QAction::triggered, this, &OGLWidget::query_act);

    queryEdges = new QAction(tr("Query edge"), this);
    queryEdges->setShortcut(QKeySequence(tr("Alt+E")));
    queryEdges->setCheckable(true);
    queryEdges->setChecked(false);
    connect(queryEdges, &QAction::triggered, this, &OGLWidget::query_act);

    queryTriangles = new QAction(tr("Query triangle"), this);
    queryTriangles->setShortcut(QKeySequence(tr("Alt+T")));
    queryTriangles->setCheckable(true);
    queryTriangles->setChecked(false);
    connect(queryTriangles, &QAction::triggered, this, &OGLWidget::query_act);

    queryTetra = new QAction(tr("Query tetrahedron"), this);
    //queryTetra->setShortcut(QKeySequence(tr("Alt+T")));
    queryTetra->setCheckable(true);
    queryTetra->setChecked(false);
    connect(queryTetra, &QAction::triggered, this, &OGLWidget::query_act);

    showqueryEdges = new QAction(tr("Show query edge"), this);
    //showqueryEdges->setShortcut(QKeySequence(tr("Alt+E")));
    showqueryEdges->setCheckable(true);
    showqueryEdges->setChecked(false);
    connect(showqueryEdges, &QAction::triggered, this, &OGLWidget::query_act);

    zoomRectAct = new QAction(tr("&Zoom Selected"), this);
    //zoomRectAct->setShortcut(QKeySequence(tr("Z")));
    zoomRectAct->setStatusTip(tr("Zoom selected"));
    zoomRectAct->setCheckable(true);
    connect(zoomRectAct, &QAction::triggered, this, &OGLWidget::reset_volume_act);

    // Mouse actions
    QActionGroup *actionQueryGroup = new QActionGroup(this);
    actionQueryGroup->addAction(queryVertices);
    actionQueryGroup->addAction(queryEdges);
    actionQueryGroup->addAction(queryTriangles);
    actionQueryGroup->addAction(queryTetra);
    actionQueryGroup->addAction(showqueryEdges);

    clearSelectionAct = new QAction(tr("Clear selection"), this);
    //clearSelectionAct->setShortcut(QKeySequence(tr("Alt+C")));
    //clearSelectionAct->setCheckable(true);
    //clearSelectionAct->setChecked(false);
    connect(clearSelectionAct, &QAction::triggered, this, &OGLWidget::clear_select_act);

    //==========================================================================
    // View options for lifting transformation
    
    viewParaboildAct = new QAction(tr("Show Paraboiloid"), this);
    //viewParaboildAct->setShortcut(QKeySequence(tr("Alt+E")));
    viewParaboildAct->setCheckable(true);
    viewParaboildAct->setChecked(false); // default
    connect(viewParaboildAct, &QAction::triggered, this, &OGLWidget::view_update_act);
    
    viewWeightsAct = new QAction(tr("Show weights"), this);
    //viewWeightsAct->setShortcut(QKeySequence(tr("Alt+E")));
    viewWeightsAct->setCheckable(true);
    viewWeightsAct->setChecked(false); // default
    connect(viewWeightsAct, &QAction::triggered, this, &OGLWidget::view_update_act);

    viewPolesAct = new QAction(tr("Lifted points (poles)"), this);
    //viewPolesAct->setShortcut(QKeySequence(tr("Alt+E")));
    viewPolesAct->setCheckable(true);
    viewPolesAct->setChecked(false); // default
    connect(viewPolesAct, &QAction::triggered, this, &OGLWidget::view_update_act);

    viewPolarsAct = new QAction(tr("Lifted dual (polars)"), this);
    //viewPolarsAct->setShortcut(QKeySequence(tr("Alt+E")));
    viewPolarsAct->setCheckable(true);
    viewPolarsAct->setChecked(false); // default
    connect(viewPolarsAct, &QAction::triggered, this, &OGLWidget::view_update_act);

    viewCircleAct = new QAction(tr("Show Orthocircle"), this);
    //viewCircleAct->setShortcut(QKeySequence(tr("Alt+E")));
    viewCircleAct->setCheckable(true);
    viewCircleAct->setChecked(false); // default
    connect(viewCircleAct, &QAction::triggered, this, &OGLWidget::view_update_act);

    viewlifttransformAct = new QAction(tr("Lifting (convex hull)"), this);
    viewlifttransformAct->setCheckable(true);
    viewlifttransformAct->setToolTip(tr("show the lift transformation"));
    viewlifttransformAct->setChecked(false); // default
    connect(viewlifttransformAct, &QAction::triggered, this, &OGLWidget::view_update_act);

    viewprojectionAct = new QAction(tr("Projection (Delaunay)"), this);
    viewprojectionAct->setCheckable(true);
    viewprojectionAct->setChecked(true); // default
    connect(viewprojectionAct, &QAction::triggered, this, &OGLWidget::view_update_act);

    viewVoronoiAct = new QAction(tr("Dual (Voronoi) diagram"), this);
    viewVoronoiAct->setCheckable(true);
    viewVoronoiAct->setChecked(false); // default
    connect(viewVoronoiAct, &QAction::triggered, this, &OGLWidget::view_update_act);

    setZliftAct = new QAction(tr("Set zlift..."), this);
    connect(setZliftAct, &QAction::triggered, this, &OGLWidget::set_zlift_act);

    // View options for lifting transformation
    //==========================================================================


    //viewfunctionAct = new QAction(tr("f"), this);
    //viewfunctionAct->setCheckable(true);
    //viewfunctionAct->setToolTip(tr("show function"));
    //viewfunctionAct->setChecked(false); // default
    //connect(viewfunctionAct, &QAction::triggered, this, &OGLWidget::view_update_act);
    //
    //viewfunctionXAct = new QAction(tr("fx"), this);
    //viewfunctionXAct->setCheckable(true);
    //viewfunctionXAct->setToolTip(tr("show fx"));
    //viewfunctionXAct->setChecked(false); // default
    //connect(viewfunctionXAct, &QAction::triggered, this, &OGLWidget::view_update_act);
    //
    //viewfunctionYAct = new QAction(tr("fy"), this);
    //viewfunctionYAct->setCheckable(true);
    //viewfunctionYAct->setToolTip(tr("show fy"));
    //viewfunctionYAct->setChecked(false); // default
    //connect(viewfunctionYAct, &QAction::triggered, this, &OGLWidget::view_update_act);
}

void OGLWidget::createMenus()
{
    QMenuBar *menubar = _menubar;

    QMenu *fileMenu = menubar->addMenu(tr("File"));
    fileMenu->addAction(openAct);
    fileMenu->addAction(saveAct);
    fileMenu->addAction(closeAct);
    fileMenu->addSeparator();

    QMenu *viewMenu = menubar->addMenu(tr("View"));
    viewMenu->addAction(rotateAct);
    viewMenu->addAction(transAct);
    viewMenu->addAction(shrinkAct);
    viewMenu->addAction(planeAct);
    //viewMenu->addAction(zoomRectAct);
    viewMenu->addAction(selectAct);

    /*
    QMenu *debugMenu = menubar->addMenu(tr("Debug"));
    debugMenu->addAction(queryVertices);
    debugMenu->addAction(queryEdges);
    debugMenu->addAction(queryTriangles);
    debugMenu->addAction(queryTetra);
    debugMenu->addAction(showqueryEdges);
    debugMenu->addSeparator();
    debugMenu->addAction(clearSelectionAct);
    */

    // Toolbar Top
    //_toolbar_top->addAction(openAct);
    //_toolbar_top->addAction(saveAct);
    //_toolbar_top->addAction(closeAct);

    // View options
    //_toolbar_top->addSeparator();
    //_toolbar_top->addAction(rotateAct);
    //_toolbar_top->addAction(transAct);
    //_toolbar_top->addSeparator();
    //_toolbar_top->addAction(enableShrink);
    //_toolbar_top->addAction(enableCutPlane);
    //_toolbar_top->addAction(showTransparent);
    //_toolbar_top->addAction(invertCutPlane);
    //_toolbar_top->addSeparator();
    //_toolbar_top->addAction(queryVertices);


    // Left toolbar
    // View options
    _toolbar_left->addAction(resetVolumeAct);

    _toolbar_left->addSeparator();
    _toolbar_left->addAction(viewParaboildAct);
    _toolbar_left->addAction(setZliftAct);

    _toolbar_left->addSeparator();
    _toolbar_left->addAction(showVertices);
    _toolbar_left->addAction(viewWeightsAct);
    _toolbar_left->addAction(viewPolesAct);
    //_toolbar_left->addAction(showFacetEdges);
    //_toolbar_left->addAction(showSegments);
    //_toolbar_left->addSeparator();
    //_toolbar_left->addAction(showVertexLabels);

    _toolbar_left->addSeparator();
    _toolbar_left->addAction(viewprojectionAct);
    _toolbar_left->addAction(viewlifttransformAct);    
    _toolbar_left->addAction(viewVoronoiAct);
    _toolbar_left->addAction(viewPolarsAct);

    _toolbar_left->addSeparator();
    _toolbar_left->addAction(viewCircleAct);
    //_toolbar_left->addAction(viewfunctionAct);
    //_toolbar_left->addAction(viewfunctionXAct);
    //_toolbar_left->addAction(viewfunctionYAct);

    _toolbar_left->addAction(showEdges);
    _toolbar_left->addAction(showFacets);
    _toolbar_left->addAction(showTransparent);

    //_toolbar_left->addSeparator();

    //_toolbar_left->addAction(rotateAct);
    //_toolbar_left->addAction(transAct);
    //_toolbar_left->addAction(shrinkAct);
    //_toolbar_left->addAction(planeAct);
    //_toolbar_left->addAction(selectAct);
    //_toolbar_left->addAction(zoomRectAct);
}

//=================================================================================
// public slots

void OGLWidget::ShowContextMenu(const QPoint& pos) // this is a slot
{
    // for most widgets
    QPoint globalPos = this->mapToGlobal(pos);
    // for QAbstractScrollArea and derived classes you would use:
    // QPoint globalPos = myWidget->viewport()->mapToGlobal(pos);

    QMenu myMenu;
    myMenu.addAction(showAxis);
    myMenu.addAction(showBbox);
    myMenu.addAction(showPlane);
    //myMenu.addAction(showTransparent);
    //myMenu.addSeparator();
    //myMenu.addAction(showSubfaces);
    //myMenu.addAction(showSubsegs);
    //myMenu.addAction(showFixedFaces);
    //myMenu.addAction(showMissingFacets);
    //myMenu.addAction(showIntersectFacets);
    //myMenu.addAction(showTestedEles);
    //myMenu.addAction(saveScreenAct);
    //QMenu* subMenuView = myMenu.addMenu("View options");
    //subMenuView->addAction(rotateAct);

    qDebug()<<"show context menu";
    QAction* selectedItem = myMenu.exec(globalPos);
    /*
    if (selectedItem)
    {
        // something was chosen, do stuff
    }
    else
    {
        // nothing was chosen
    }
    */
}

void OGLWidget::openFile()
{
    /*
    // Read input from file(s).
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open File"),
            _file_path,
            tr("Geometry & meshes (*.node *.ele *.face *.edge *.smesh *.mesh *.ply *.off *.stl)"));

    //QByteArray ba = fileName.toLatin1();
    QByteArray ba = fileName.toLocal8Bit();
    const char *c_str2 = ba.data();
    if (c_str2!=NULL) {
      _file_path = QFileInfo(fileName).path(); // store path for next time
      qDebug()<<" Path="<<_file_path<<", name="<<fileName;

      if (_mesh != NULL) delete _mesh;
      _mesh = new Triangulation();

      char programname[100];
      char filename[256];
      strcpy(programname, "tetgen2");
      strcpy(filename, c_str2);

      int argc = 2;
      char *argv[2];
      argv[0] = programname;
      argv[1] = filename;

      _mesh->parse_commandline(argc, argv);

      if (_mesh->read_mesh()) {
        // mesh read, show it.
        if (_mesh->tr_tets != NULL) {
          _mesh->reconstruct_mesh();
        }
        QString text = "Open file "+fileName+" successfully!";
        m_statusRight->setText(text);
        plot3d();
        update();
      } else {
        QMessageBox msg;
        QString text = QString("Fail to read geometry/mesh from file ") + fileName;
        msg.setText(text);
        msg.exec();
        delete _mesh;
        _mesh = NULL;
      }
    }
    */
}

void OGLWidget::saveFile()
{

}

void OGLWidget::closeFile()
{

}

void OGLWidget::open_preferDialog()
{

}

void OGLWidget::mouse_act()
{
    if (rotateAct->isChecked()) {
      _mouse_act = eRot;
    } else if (transAct->isChecked()) {
      _mouse_act = eTrans;
    } else if (planeAct->isChecked()) {
      _mouse_act = eCutPlane;
    } else if (shrinkAct->isChecked()) {
      _mouse_act = eShrink;
    } else if (selectAct->isChecked()) {
      _mouse_act = eSelect;
       // clear selection.
       _pos.init(); _ray.init();
    }
}

void OGLWidget::query_act()
{
    /*
    if (_mesh == NULL) {
      QMessageBox msg;
      msg.setText(tr("No mesh!"));
      msg.exec();
      return;
    }

    bool ok;

    if (queryVertices->isChecked()) {
      int min_value = _mesh->io_firstindex;
      int max_value = _mesh->ct_in_vrts;
      int default_value = min_value;
      int i = QInputDialog::getInt(this, // QWidget
                                   tr("Query vertices"), // title
                                   tr("vertex index"), // label
                                   default_value, min_value, max_value, 1, &ok);
      if (ok) {
        Vertex *v = NULL;
        if (_mesh->get_vrt(i, &v)) {
          * (Vertex *) _sel_vrts->alloc() = *v;
          update(); // draw the selected objects.
        } else {
          QMessageBox msg; msg.setText("The given vertex does not exist!");
          msg.exec();
        }
      }
    } else if (queryEdges->isChecked()) {
      QString text = QInputDialog::getText(this, // Qwidget
                                           tr("Query edges"), // title
                                           tr("Start, end"),  // label
                                           QLineEdit::Normal, QString(""), &ok);
      if (ok) {
        if (!text.isEmpty()) {
          //QByteArray ba = fileName.toLatin1();
          QByteArray ba = text.toLocal8Bit();
          const char *c_str2 = ba.data();
          int e1=0, e2=0;
          sscanf(c_str2, "%d,%d", &e1, &e2);
          printf("  select edge (%d,%d)\n", e1, e2);
          TEdge E;
          if (_mesh->get_edge(e1, e2, E)) {
            Vertex *vv = (Vertex *) _sel_edges->alloc();
            vv[0] = *(E.org());
            vv[1] = *(E.dest());
            QString tt = "Selected edge "+QString::number(e1)+","+QString::number(e2);
            m_statusRight->setText(tt);
            update();
          } else {
            QMessageBox msg; msg.setText("The given Edge does not exist!");
            msg.exec();
          }
        }
      }
    } else if (showqueryEdges->isChecked()) {
      QString text = QInputDialog::getText(this, // Qwidget
                                           tr("Add query edges"), // title
                                           tr("Start, end"),  // label
                                           QLineEdit::Normal, QString(""), &ok);
      if (ok) {
        if (!text.isEmpty()) {
          QByteArray ba = text.toLocal8Bit();
          const char *c_str2 = ba.data();
          int e1=0, e2=0;
          sscanf(c_str2, "%d,%d", &e1, &e2);
          printf("  select edge (%d,%d)\n", e1, e2);
          Vertex *v1, *v2;
          if (_mesh->get_vrt(e1, &v1)) {
            if (_mesh->get_vrt(e2, &v2)) {
              Vertex *vv = (Vertex *) _sel_edges->alloc();
              vv[0] = *v1;
              vv[1] = *v2;
              QString tt = "Added query edge "+QString::number(e1)+","+QString::number(e2);
              m_statusRight->setText(tt);
              update();
            } else {
              QString tt = "Vertex "+QString::number(e2)+" does not exist.";
              QMessageBox msg; msg.setText(tt);
              msg.exec();
            }
          } else {
            QString tt = "Vertex "+QString::number(e1)+" does not exist.";
            QMessageBox msg; msg.setText(tt);
            msg.exec();
          }
        }
      }
    }
    */
}

void OGLWidget::clear_select_act()
{
    _sel_vrts->clean();
    _sel_edges->clean();

    //QMessageBox msg;
    //msg.setText("Selection cleared.");
    //msg.exec();

    update();
}

void OGLWidget::reset_volume_act()
{
    reset_volume();
    update();
}

void OGLWidget::view_update_act()
{
    _object_valid = 0; // re-draw the object.
    update();
}

void OGLWidget::set_zlift_act()
{    
    char str[256];
    sprintf(str, "zmin=%g, zmax=%g", _min[2], _max[2]);
    QString hint = str;

    bool ok;
    //double d = QInputDialog::getDouble(this, tr("Vertical shift distance"),
    //                                   hint, _zshift, -10000., 10000., 3, &ok);
    QString text = QInputDialog::getText(this, // Qwidget
                                               tr("Vertical shift distance"), // title
                                               hint,  // label
                                               QLineEdit::Normal, QString::number(_zshift), &ok);

    if (ok && !text.isEmpty()) {
      _zshift = text.toDouble();
      //reset_volume();
      _object_valid = 0; // re-draw the object.
      update();
    }
}

void OGLWidget::update_act()
{
    update();
}

//=================================================================================

// prepare_view_transformation()
// {
//     rotx = roty = rotz = 0.;
//
//     rotz = _rotY;
//     rotx = _rotX;
//
//    glRotated(rotx,1.0,0.0,0.0);
//    glRotated(roty,0.0,1.0,0.0);
//    glRotated(rotz,0.0,0.0,1.0);
//
// }

void OGLWidget::keyPressEvent(QKeyEvent *event)
{
    if (event->key() == Qt::Key_Escape) {
      qDebug()<<"ESC is pressed";
    } else if (event->key() == Qt::Key_8) {
      //qDebug()<<"Key 8 is pressed";
      //_plane_dir = 2; // (-Z, X-Y plane)
      // see from -Z to +Z direction (0,0,1) X-Y plane
      // - rotate scene along +Y direction 180 degree, using the right
      //   hand rule (right thumb point to +Z).
      _rotX = 0.0;
      _rotY = 0.0;  // rotz
      _rotZ = 180.; // roty
      update();
    } else if (event->key() == Qt::Key_7) {
      //qDebug()<<"Key 7 is pressed";
      //_plane_dir = 2; // (-Z, X-Y plane)
      // see from +Z to -Z direction (0,0,-1) X-Y plane
      _rotX = _rotY = 0.0;
      _rotZ = 0.; // roty
      update();
    } else if (event->key() == Qt::Key_5) {
      //qDebug()<<"Key 5 is pressed";
      _ortho = !_ortho;
      update();
    } else if (event->key() == Qt::Key_4) {
      //_plane_dir = 1; // (+Y, X-Z plane)
      // see from +Y to -Y direction (0,-1,0) X-Z plane
      // This is achieved by two steps:
      // - rotate scene along +X direction 90 degree, using the left
      //   hand rule (left thumb point to +X).
      // - rotate scene along +Z direction 180 degree, using the right
      //   hand rule (right thumb point to +Z).
      _rotX = -90.;
      _rotY = 180.;
      _rotZ = 0.; // roty
      update();
    } else if (event->key() == Qt::Key_3) {
      //_plane_dir = 1; // (-Y, X-Z plane)
      // see from -Y to +Y direction (0,1,0) X-Z plane
      // rotate scene along +X direction 90 degree, using the left
      //   hand rule (left thumb point to +X).
      _rotX = -90.;
      _rotY = 0.;
      _rotZ = 0.; // roty
      update();
    } else if (event->key() == Qt::Key_2) {
      //_plane_dir = 0; // (X, Y-Z plane)
      // see from +X to -X direction (-1,0,0) Y-Z plane
      // This is achieved by two steps:
      // - rotate scene along +X direction 90 degree, using the left
      //   hand rule (left thumb point to +X).
      // - rotate scene along +Z direction 90 degree, using the right
      //   hand rule (right thumb point to +Z).
      _rotX = -90.;
      _rotY = -90.;
      _rotZ = 0.; // roty
      update();
    } else if (event->key() == Qt::Key_1) {
      //_plane_dir = 0; // (X, Y-Z plane)
      // see from -X to +X direction (1,0,0)  Y-Z plane
      // This is achieved by two steps:
      // - rotate scene along +X direction 90 degree, using the left
      //   hand rule (left thumb point to +X).
      // - rotate scene along +Z direction 90 degree, using the right
      //   hand rule (right thumb point to +Z).
      _rotX = -90.;
      _rotY = 90.; //90.;
      _rotZ = 0.; // roty
      update();
    } else if (event->key() == Qt::Key_0) {
      //qDebug()<<"Key 0 (reset) is pressed";
      reset_view_transformations();
      update();
    } else if (event->key() == Qt::Key_X) {
      if (event->modifiers() & Qt::AltModifier) {
        //qDebug()<<"Alt+X is pressed";
        _plane_dir = 0;
        if (enableCutPlane->isChecked()) {
          _object_valid = 0;
        }
        update();
      }
    } else if (event->key() == Qt::Key_Y) {
      if (event->modifiers() & Qt::AltModifier) {
        //qDebug()<<"Alt+Y is pressed";
        _plane_dir = 1;
        if (enableCutPlane->isChecked()) {
          _object_valid = 0;
        }
        update();
      }
    } else if (event->key() == Qt::Key_Z) {
      if (event->modifiers() & Qt::AltModifier) {
        //qDebug()<<"Alt+Z is pressed";
        _plane_dir = 2;
        if (enableCutPlane->isChecked()) {
          _object_valid = 0;
        }
        update();
      }
    }
}

void OGLWidget::mousePressEvent(QMouseEvent *e)
{
    _prevX = e->x();
    _prevY = e->y();

    if (_mouse_act == eSelect) {
      double wx, wy, wz;
      unproject(_prevX, _prevY, &wx, &wy, &wz);
      update();
    }
}

void OGLWidget::mouseReleaseEvent(QMouseEvent *)
{
}

void OGLWidget::mouseMoveEvent(QMouseEvent *e)
{
    Qt::MouseButtons btns = e->buttons();

    //std::cout << e->x() << " " << e->y() << std::endl;

    if (btns == Qt::LeftButton) {
      if (_mouse_act == eRot) {
        if (e->x() > _prevX)
          _rotY += 2.; // 6.; // /_zoom;
        else if (e->x() < _prevX)
          _rotY -= 2.; // 6.; // /_zoom;
        if (e->y() > _prevY)
          _rotX += 2.; // 6.; // /_zoom;
        else if (e->y() < _prevY)
          _rotX -= 2.; // 6.; // /_zoom;
        update();
      } else if (_mouse_act == eTrans) {
        if (e->x() > _prevX)
          _tranX += 0.03;
        else if (e->x() < _prevX)
          _tranX -= 0.03;
        if (e->y() > _prevY)
          _tranY -= 0.03;
        else if (e->y() < _prevY)
          _tranY += 0.03;
        update();
      } else if (_mouse_act == eCutPlane) {
        // get the maximum movement between X and Y
        int dx = e->x() - _prevX;
        int dy = e->y() < _prevY;
        int d = dx;
        if (abs(dx) < abs(dy)) d = dy;
        _plane_d += ((double) d) / 100.;
        qDebug()<<"Move CutPlane: d="<<d<<", q/10."<<((double) d) / 10.;
        if (_plane_d < 0.) _plane_d = 0.;
        if (_plane_d > 1.) _plane_d = 1.0;
        if (enableCutPlane->isChecked()) { // if cutting is enabled.
          _object_valid=0;
        }
        update();
      } else if (_mouse_act == eShrink) {
        if (enableShrink->isChecked()) {
          if (e->y() > _prevY) {
            _shrink -= 0.03;
            if (_shrink < 0.01) _shrink = 0.01;
            //_object_valid=0;
            //update();
            view_update_act();
          } else if (e->y() < _prevY) {
            _shrink += 0.03;
            if (_shrink > 0.99) _shrink = 0.99;
            //_object_valid=0;
            //update();
            view_update_act();
          }
        }
      }

      _prevX = e->x();
      _prevY = e->y();
    }
    /*
    else if (btns == Qt::RightButton)
    {

    }
    */
}

void OGLWidget::wheelEvent(QWheelEvent *e)
{
    if (e->delta() > 0)
        _zoom += 0.05*_zoom;
    else if (e->delta() < 0)
        if (_zoom > 0.2)
            _zoom -= 0.05*_zoom;

    if (_zoom < 1/_zoom_max)
        _zoom = 1/_zoom_max;

    //if (_zoom > _zoom_max)
    //    _zoom = _zoom_max;

    e->accept();

    update();
}

//=====================================================================================

void OGLWidget::projection(double Ox, double Oy, double Oz, int* Wx, int* Wy)
{
    char sbuff[128];
    if (0) {
      sprintf(sbuff, " %g,%g,%g,1.0", Ox, Oy, Oz);
      qDebug()<<"  object coordinates:"<<sbuff;
    }

    //for (int i = 0; i < 16; i++) _Mv[i] = 0.;
    //glGetFloatv(GL_MODELVIEW_MATRIX, _Mv);

    // eye coordinates
    double Pe[4];
    Pe[0] = _Mv[0]*Ox+_Mv[4]*Oy+_Mv[ 8]*Oz+_Mv[12]*1.0;
    Pe[1] = _Mv[1]*Ox+_Mv[5]*Oy+_Mv[ 9]*Oz+_Mv[13]*1.0;
    Pe[2] = _Mv[2]*Ox+_Mv[6]*Oy+_Mv[10]*Oz+_Mv[14]*1.0;
    Pe[3] = _Mv[3]*Ox+_Mv[7]*Oy+_Mv[11]*Oz+_Mv[15]*1.0;

    if (0) {
      sprintf(sbuff, " %g,%g,%g,%g", Pe[0], Pe[1], Pe[2], Pe[3]);
      qDebug()<<"  eye coordinates:"<<sbuff;
    }

    //for (int i = 0; i < 16; i++) _Mp[i] = 0.;
    //glGetFloatv(GL_PROJECTION_MATRIX, _Mp);

    // (projection) clip coordinates
    double Pc[4];
    Pc[0] = _Mp[0]*Pe[0]+_Mp[4]*Pe[1]+_Mp[ 8]*Pe[2]+_Mp[12]*Pe[3];
    Pc[1] = _Mp[1]*Pe[0]+_Mp[5]*Pe[1]+_Mp[ 9]*Pe[2]+_Mp[13]*Pe[3];
    Pc[2] = _Mp[2]*Pe[0]+_Mp[6]*Pe[1]+_Mp[10]*Pe[2]+_Mp[14]*Pe[3];
    Pc[3] = _Mp[3]*Pe[0]+_Mp[7]*Pe[1]+_Mp[11]*Pe[2]+_Mp[15]*Pe[3];

    if (0) {
      sprintf(sbuff, " %g,%g,%g,%g", Pc[0], Pc[1], Pc[2], Pc[3]);
      qDebug()<<"  clip coordinates:"<<sbuff;
    }

    if (!_ortho) {
      for (int i = 0; i < 3; i++) Pc[i] /= Pc[3];
      if (0) {
        sprintf(sbuff, " %g,%g,%g,%g", Pc[0], Pc[1], Pc[2], Pc[3]);
        qDebug()<<"  NDC coordinates:"<<sbuff;
      }
    }

    // window coordinates
    // glViewport(0, 0, w, h);
    // glDepthRange(n,f);
    double w = (double) _w, h = (double) _h;
    if (0) {
      sprintf(sbuff, "View port (0, 0, w(%d), h(%d)):", (int) w, (int) h); qDebug()<<sbuff;
    }

    int Pw[2];
    Pw[0] =      (int) (w/2. * Pc[0] + w/2.);
    Pw[1] =  h - (int) (h/2. * Pc[1] + h/2.);

    if (0) {
      sprintf(sbuff, " %d,%d", Pw[0], Pw[1]);
      qDebug()<<"  window coordinates:"<<sbuff;
      qDebug()<<"...";
    }

    *Wx = Pw[0];
    *Wy = Pw[1];
}

// http://nehe.gamedev.net/article/using_gluunproject/16013/
// gluUnProject converts Windows screen coordinates to OpenGL coordinates.
// This means that you can get the position of your mouse on an OpenGL Window
// and use this method to find the x, y and z coordinate of where you clicked.
//
// [stackoverflow]
// // https://stackoverflow.com/questions/3089271/converting-mouse-position-to-world-position-opengl
// You must enable depth buffering for the code in that article to work.
// The Z value for mouse coordinates is determined based on the value in the
// depth buffer at that point. In your initialization code, make sure you do
// the following:
//   glEnable(GL_DEPTH);

void OGLWidget::unproject(int Wx, int Wy, double *Ox, double *Oy, double *Oz)
{
    GLint viewport[4];
    GLdouble modelview[16];
    GLdouble projection[16];
    GLfloat winX, winY, winZ;
    GLdouble posX, posY, posZ;

    //glGetDoublev( GL_MODELVIEW_MATRIX, modelview );
    for (int i = 0; i < 16; i++) modelview[i] = (GLdouble) _Mv[i];
    //glGetDoublev( GL_PROJECTION_MATRIX, projection );
    for (int i = 0; i < 16; i++) projection[i] = (GLdouble) _Mp[i];
    //glGetIntegerv( GL_VIEWPORT, viewport );
    viewport[0] = viewport[1] = 0.;
    viewport[2] = _w;
    viewport[3] = _h;

    qDebug()<<"viewport: ["<<viewport[0]<<","<<viewport[1]<<","<<viewport[2]<<","<<viewport[3]<<"]";

    winX = (float)Wx;
    winY = (float)viewport[3] - (float)Wy;
    glReadPixels( Wx, int(winY), 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &winZ );

    gluUnProject( winX, winY, winZ, modelview, projection, viewport, &posX, &posY, &posZ);
    char str[128];
    sprintf(str, "(%d,%d,%d)->(%g,%g,%g)", Wx, Wy, (int)winZ, posX, posY, posZ);
    qDebug()<<str;

    *Ox = _pos.crd[0] = posX;
    *Oy = _pos.crd[1] = posY;
    *Oz = _pos.crd[2] = posZ;
    _pos.tag = 1;

    // Experimental: try to get a ray
    raycasting(Wx, Wy, &posX, &posY, &posZ);
    sprintf(str, "(%d,%d)->(%g,%g,%g)", Wx, Wy, posX, posY, posZ);
    qDebug()<<str;

    _ray.crd[0] = posX;
    _ray.crd[1] = posY;
    _ray.crd[2] = posZ;
}

// Inverting a 4x4 matrix which is given in array of 16 floats in cloumn-major order.
// https://stackoverflow.com/questions/1148309/inverting-a-4x4-matrix
bool InvertMatrix(const double m[16], double invOut[16])
{
    double inv[16], det;
    int i;

    inv[0] = m[5]  * m[10] * m[15] -
             m[5]  * m[11] * m[14] -
             m[9]  * m[6]  * m[15] +
             m[9]  * m[7]  * m[14] +
             m[13] * m[6]  * m[11] -
             m[13] * m[7]  * m[10];

    inv[4] = -m[4]  * m[10] * m[15] +
              m[4]  * m[11] * m[14] +
              m[8]  * m[6]  * m[15] -
              m[8]  * m[7]  * m[14] -
              m[12] * m[6]  * m[11] +
              m[12] * m[7]  * m[10];

    inv[8] = m[4]  * m[9] * m[15] -
             m[4]  * m[11] * m[13] -
             m[8]  * m[5] * m[15] +
             m[8]  * m[7] * m[13] +
             m[12] * m[5] * m[11] -
             m[12] * m[7] * m[9];

    inv[12] = -m[4]  * m[9] * m[14] +
               m[4]  * m[10] * m[13] +
               m[8]  * m[5] * m[14] -
               m[8]  * m[6] * m[13] -
               m[12] * m[5] * m[10] +
               m[12] * m[6] * m[9];

    inv[1] = -m[1]  * m[10] * m[15] +
              m[1]  * m[11] * m[14] +
              m[9]  * m[2] * m[15] -
              m[9]  * m[3] * m[14] -
              m[13] * m[2] * m[11] +
              m[13] * m[3] * m[10];

    inv[5] = m[0]  * m[10] * m[15] -
             m[0]  * m[11] * m[14] -
             m[8]  * m[2] * m[15] +
             m[8]  * m[3] * m[14] +
             m[12] * m[2] * m[11] -
             m[12] * m[3] * m[10];

    inv[9] = -m[0]  * m[9] * m[15] +
              m[0]  * m[11] * m[13] +
              m[8]  * m[1] * m[15] -
              m[8]  * m[3] * m[13] -
              m[12] * m[1] * m[11] +
              m[12] * m[3] * m[9];

    inv[13] = m[0]  * m[9] * m[14] -
              m[0]  * m[10] * m[13] -
              m[8]  * m[1] * m[14] +
              m[8]  * m[2] * m[13] +
              m[12] * m[1] * m[10] -
              m[12] * m[2] * m[9];

    inv[2] = m[1]  * m[6] * m[15] -
             m[1]  * m[7] * m[14] -
             m[5]  * m[2] * m[15] +
             m[5]  * m[3] * m[14] +
             m[13] * m[2] * m[7] -
             m[13] * m[3] * m[6];

    inv[6] = -m[0]  * m[6] * m[15] +
              m[0]  * m[7] * m[14] +
              m[4]  * m[2] * m[15] -
              m[4]  * m[3] * m[14] -
              m[12] * m[2] * m[7] +
              m[12] * m[3] * m[6];

    inv[10] = m[0]  * m[5] * m[15] -
              m[0]  * m[7] * m[13] -
              m[4]  * m[1] * m[15] +
              m[4]  * m[3] * m[13] +
              m[12] * m[1] * m[7] -
              m[12] * m[3] * m[5];

    inv[14] = -m[0]  * m[5] * m[14] +
               m[0]  * m[6] * m[13] +
               m[4]  * m[1] * m[14] -
               m[4]  * m[2] * m[13] -
               m[12] * m[1] * m[6] +
               m[12] * m[2] * m[5];

    inv[3] = -m[1] * m[6] * m[11] +
              m[1] * m[7] * m[10] +
              m[5] * m[2] * m[11] -
              m[5] * m[3] * m[10] -
              m[9] * m[2] * m[7] +
              m[9] * m[3] * m[6];

    inv[7] = m[0] * m[6] * m[11] -
             m[0] * m[7] * m[10] -
             m[4] * m[2] * m[11] +
             m[4] * m[3] * m[10] +
             m[8] * m[2] * m[7] -
             m[8] * m[3] * m[6];

    inv[11] = -m[0] * m[5] * m[11] +
               m[0] * m[7] * m[9] +
               m[4] * m[1] * m[11] -
               m[4] * m[3] * m[9] -
               m[8] * m[1] * m[7] +
               m[8] * m[3] * m[5];

    inv[15] = m[0] * m[5] * m[10] -
              m[0] * m[6] * m[9] -
              m[4] * m[1] * m[10] +
              m[4] * m[2] * m[9] +
              m[8] * m[1] * m[6] -
              m[8] * m[2] * m[5];

    det = m[0] * inv[0] + m[1] * inv[4] + m[2] * inv[8] + m[3] * inv[12];

    if (det == 0)
        return false;

    det = 1.0 / det;

    for (i = 0; i < 16; i++)
        invOut[i] = inv[i] * det;

    /*
    // Debug
    //vec4 ray_eye = inverse(projection_matrix) * ray_clip;
    double IM[16];
    for (i = 0; i < 16; i++) IM[i] = 0.;
    IM[0]  = invOut[0]*m[0] +invOut[4]*m[1] +invOut[ 8]*m[2] +invOut[12]*m[3];
    IM[1]  = invOut[1]*m[0] +invOut[5]*m[1] +invOut[ 9]*m[2] +invOut[13]*m[3];
    IM[2]  = invOut[2]*m[0] +invOut[6]*m[1] +invOut[10]*m[2] +invOut[14]*m[3];
    IM[3]  = invOut[3]*m[0] +invOut[7]*m[1] +invOut[11]*m[2] +invOut[15]*m[3];

    IM[4]  = invOut[0]*m[4] +invOut[4]*m[5] +invOut[ 8]*m[6] +invOut[12]*m[7];
    IM[5]  = invOut[1]*m[4] +invOut[5]*m[5] +invOut[ 9]*m[6] +invOut[13]*m[7];
    IM[6]  = invOut[2]*m[4] +invOut[6]*m[5] +invOut[10]*m[6] +invOut[14]*m[7];
    IM[7]  = invOut[3]*m[4] +invOut[7]*m[5] +invOut[11]*m[6] +invOut[15]*m[7];

    IM[8]  = invOut[0]*m[8] +invOut[4]*m[9] +invOut[ 8]*m[10] +invOut[12]*m[11];
    IM[9]  = invOut[1]*m[8] +invOut[5]*m[9] +invOut[ 9]*m[10] +invOut[13]*m[11];
    IM[10] = invOut[2]*m[8] +invOut[6]*m[9] +invOut[10]*m[10] +invOut[14]*m[11];
    IM[11] = invOut[3]*m[8] +invOut[7]*m[9] +invOut[11]*m[10] +invOut[15]*m[11];

    IM[12] = invOut[0]*m[12]+invOut[4]*m[13]+invOut[ 8]*m[14] +invOut[12]*m[15];
    IM[13] = invOut[1]*m[12]+invOut[5]*m[13]+invOut[ 9]*m[14] +invOut[13]*m[15];
    IM[14] = invOut[2]*m[12]+invOut[6]*m[13]+invOut[10]*m[14] +invOut[14]*m[15];
    IM[15] = invOut[3]*m[12]+invOut[7]*m[13]+invOut[11]*m[14] +invOut[15]*m[15];

    char sbuff[128];
    qDebug()<<"Verifying inverting matrix:";
    sprintf(sbuff, " | %g  %g  %g  %g |", IM[0], IM[4], IM[ 8], IM[12]); qDebug()<<sbuff;
    sprintf(sbuff, " | %g  %g  %g  %g |", IM[1], IM[5], IM[ 9], IM[13]); qDebug()<<sbuff;
    sprintf(sbuff, " | %g  %g  %g  %g |", IM[2], IM[6], IM[10], IM[14]); qDebug()<<sbuff;
    sprintf(sbuff, " | %g  %g  %g  %g |", IM[3], IM[7], IM[11], IM[15]); qDebug()<<sbuff;
    qDebug()<<"....";
    */

    return true;
}

// http://antongerdelan.net/opengl/raycasting.html
// Mouse Picking with Ray Casting, by Anton Gerdelan.
// (with my own correction for both Projection and Perspective transformations).
// (the Windows Z depth coordinate is needed).
void OGLWidget::raycasting(int mouse_x, int mouse_y, double *Rx, double *Ry, double *Rz)
{
    GLdouble model[16], inv_model[16];
    GLdouble proj[16], inv_proj[16];

    printf("Ray casting from mouse position (%d, %d)\n", mouse_x, mouse_y);

    // Get the NDC coordinates for (mouse_x, mouse_y), see (see notes of Song Ho Ahn).
    float x = (2.0f * mouse_x) / _w - 1.0f;
    float y = 1.0f - (2.0f * mouse_y) / _h;
    float z = 1.0f; // it is not decided yet (only needed for orthogonal projection).

    if (_ortho) {
      // Get the Windows Z coordinate (need Depth Buffer, glEnable(GL_DEPTH)).
      GLfloat winX, winY, winZ;
      winX = (float)mouse_x;
      winY = _h - (float)mouse_y;
      //glReadPixels( Wx, int(winY), 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &winZ );
      glReadPixels( mouse_x, int(winY), 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &winZ );
      // Get the NDC coordinate for z (see notes of Song Ho Ahn).
      z = (winZ - (_zFar+_zNear)/2.) * 2./ (_zFar - _zNear);
    }

    //vec3 ray_nds = vec3(x, y, z);
    //double ray_nds[3] = {x, y, z};
    double ray_nds[3];
    ray_nds[0] = x;
    ray_nds[1] = y;
    ray_nds[2] = z;
    printf("  nds coords: (%g, %g, %g)\n", ray_nds[0], ray_nds[1], ray_nds[2]);

    //vec4 ray_clip = vec4(ray_nds.xy, -1.0, 1.0);
    //double ray_clip[4] = {ray_nds[0], ray_nds[1], -1.0, 1.0};
    double ray_clip[4];
    ray_clip[0] = ray_nds[0];
    ray_clip[1] = ray_nds[1];
    if (_ortho) {
      ray_clip[0] *= z;
      ray_clip[1] *= z;
    }
    ray_clip[2] = -1.0;
    ray_clip[3] =  0.0;
    printf("  clip coords: (%g, %g, %g, %g)\n", ray_clip[0], ray_clip[1], ray_clip[2], ray_clip[3]);

    //vec4 ray_eye = inverse(projection_matrix) * ray_clip;
    //glGetDoublev( GL_PROJECTION_MATRIX, projection );
    for (int i = 0; i < 16; i++) proj[i] = (GLdouble) _Mp[i];
    InvertMatrix(proj, inv_proj);
    double ray_eye[4];
    ray_eye[0] = inv_proj[0]*ray_clip[0]+inv_proj[4]*ray_clip[1]+inv_proj[ 8]*ray_clip[2]+inv_proj[12]*ray_clip[3];
    ray_eye[1] = inv_proj[1]*ray_clip[0]+inv_proj[5]*ray_clip[1]+inv_proj[ 9]*ray_clip[2]+inv_proj[13]*ray_clip[3];
    ray_eye[2] = inv_proj[2]*ray_clip[0]+inv_proj[6]*ray_clip[1]+inv_proj[10]*ray_clip[2]+inv_proj[14]*ray_clip[3];
    ray_eye[3] = inv_proj[3]*ray_clip[0]+inv_proj[7]*ray_clip[1]+inv_proj[11]*ray_clip[2]+inv_proj[15]*ray_clip[3];

    // ray_eye = vec4(ray_eye.xy, -1.0, 0.0);
    //if (_ortho) {
    //  ray_eye[0] *= z;
    //  ray_eye[1] *= z;
    //}
    ray_eye[2] = -1.0;
    ray_eye[3] =  0.0;
    printf("  eye coords: (%g, %g, %g, %g)\n", ray_eye[0], ray_eye[1], ray_eye[2], ray_eye[3]);

    //vec3 ray_wor = (inverse(view_matrix) * ray_eye).xyz;
    //glGetDoublev( GL_MODELVIEW_MATRIX, modelview );
    for (int i = 0; i < 16; i++) model[i] = (GLdouble) _Mv[i];
    InvertMatrix(model, inv_model);
    double ray_wor[4];
    ray_wor[0] = inv_model[0]*ray_eye[0]+inv_model[4]*ray_eye[1]+inv_model[ 8]*ray_eye[2]+inv_model[12]*ray_eye[3];
    ray_wor[1] = inv_model[1]*ray_eye[0]+inv_model[5]*ray_eye[1]+inv_model[ 9]*ray_eye[2]+inv_model[13]*ray_eye[3];
    ray_wor[2] = inv_model[2]*ray_eye[0]+inv_model[6]*ray_eye[1]+inv_model[10]*ray_eye[2]+inv_model[14]*ray_eye[3];
    ray_wor[3] = inv_model[3]*ray_eye[0]+inv_model[7]*ray_eye[1]+inv_model[11]*ray_eye[2]+inv_model[15]*ray_eye[3];

    ray_wor[3] = 0.0; // not used.
    printf("  world coords: (%g, %g, %g)\n", ray_wor[0], ray_wor[1], ray_wor[2]);

    // // don't forget to normalise the vector at some point
    // ray_wor = normalise(ray_wor);
    double nm = sqrt(ray_wor[0]*ray_wor[0]+ray_wor[1]*ray_wor[1]+ray_wor[2]*ray_wor[2]);
    ray_wor[0] /= nm;
    ray_wor[1] /= nm;
    ray_wor[2] /= nm;
    printf("  normalised: (%g, %g, %g)\n", ray_wor[0], ray_wor[1], ray_wor[2]);

    *Rx = ray_wor[0];
    *Ry = ray_wor[1];
    *Rz = ray_wor[2];
}

void OGLWidget::renderText(QPainter *painter, double Ox, double Oy, double Oz, const QString text)
{
    //char str[100];
    //sprintf(str, "%.2g,%.2g,%.2g", P[0], P[1], P[2]);
    int Wx, Wy;
    projection(Ox, Oy, Oz, &Wx, &Wy);
    painter->drawText(Wx, Wy, text);
}

// This function must be called after reset_volume()
void OGLWidget::triangulate_disc()
{
    if (_disc != NULL) delete _disc;
    _disc = new Triangulation();

    ScribbleArea *SA = (ScribbleArea *) _scribbleArea;
    Triangulation *_geom = SA->_geom;
    double ctx, cty, radius;
    int segnum = 72;

    ctx = 0.5 * (_geom->io_xmin + _geom->io_xmax);
    cty = 0.5 * (_geom->io_ymin + _geom->io_ymax);
    radius = 0.5 * sqrt((_geom->io_xmax-_geom->io_xmin)*(_geom->io_xmax-_geom->io_xmin) +
                        (_geom->io_ymax-_geom->io_ymin)*(_geom->io_ymax-_geom->io_ymin));

    // Generate the points on circle.
    double ang_start = 0.; // ang1 / 180. * detri2::PI;
    double dfi = 360. / segnum; // fabs(ang2 - ang1) / segnum;
    dfi = dfi / 180.0 * detri2::PI;

    /*
    int np = segnum + 1;
    double *ptlist = new double[np * 2];
    for (int i = 0; i < np; i++) {
      double *pp = &(ptlist[i*2]);
      pp[0] = radius * cos(ang_start + dfi * (double) i) + ctx;
      pp[1] = radius * sin(ang_start + dfi * (double) i) + cty;
    }
    */

    int firstindex = 1; // Set it to be 0 or 1.
    _disc->io_firstindex = firstindex;

    // Read the input points and weights.
    int i;
    double x, y;
    int nv = segnum; //_geom->tr_steiners->objects;
    _disc->ct_in_vrts = nv;
    _disc->in_vrts = new Vertex[nv];
    for (i = 0; i < nv; i++) {
      Vertex *vrt = &(_disc->in_vrts[i]);
      vrt->init();
      x = vrt->crd[0] = radius * cos(ang_start + dfi * (double) i) + ctx;;
      y = vrt->crd[1] = radius * sin(ang_start + dfi * (double) i) + cty;
      vrt->crd[2] = x*x+y*y;
      vrt->idx = i + firstindex;
    }
    _disc->ct_unused_vrts=nv;
    _disc->io_xmin = _min[0];
    _disc->io_xmax = _max[1];
    _disc->io_ymin = _min[1];
    _disc->io_ymax = _max[1];
    _disc->io_diagonal = 2. * radius;

    _disc->incremental_delaunay();

    /*
    int *elist = new int[segnum * 2];
    for (int i = 0; i < segnum; i++) {
      elist[i*2] = i;
      elist[i*2+1] = i+1;
    }
    */

    _disc->tr_segs = new arraypool(sizeof(Triang), 8);
    for (i = 0; i < segnum-1; i++) {
      Triang *seg =  (Triang *) _disc->tr_segs->alloc();
      seg->init();
      seg->vrt[0] = &(_disc->in_vrts[i]);
      seg->vrt[1] = &(_disc->in_vrts[i+1]);
      seg->tag = -1;
      //seg->val = gseg->val;
      //printf("  get segment %d,%d, tag=%d, val=%g\n", seg->vrt[0]->idx, seg->vrt[1]->idx, seg->tag, seg->val);
    }
    // The last segment.
    Triang *seg =  (Triang *) _disc->tr_segs->alloc();
    seg->init();
    seg->vrt[0] = &(_disc->in_vrts[i]);
    seg->vrt[1] = &(_disc->in_vrts[0]);
    seg->tag = -1;

    _disc->recover_segments();

    // Refine mesh with min edge length.
    _disc->op_target_length = _disc->get_distance(seg->vrt[0], seg->vrt[1]);
    _disc->op_cosminangle = cos(30.0 / 180.0 * PI);
    _disc->delaunay_refinement();

    // debug only
    //strcpy(_disc->io_outfilename, "/Users/si/tmp/disc_mesh");
    //_disc->save_triangulation();
}

// This function must be called after reset_volume()
void OGLWidget::triangulate_ellipse()
{
    if (_disc != NULL) delete _disc;
    _disc = new Triangulation();

    ScribbleArea *SA = (ScribbleArea *) _scribbleArea;
    Triangulation *_geom = SA->_geom;

    //====================================================
    // BAMG, Metric.cpp, MatVVP2x2::MatVVP2x2(const MetricAnIso M)
    double lambda1, lambda2; // eigenvalues
    double vec1[2], vec2[2]; // eigenvectors
    get_MatVVP2x2(SA->_a11, SA->_a21, SA->_a22,
                         &lambda1, &lambda2, vec1, vec2);
    /*
    {
      double a11=SA->_a11; //SA->_coef_a;
      double a21=SA->_a21; //SA->_coef_b;
      double a22=SA->_a22; //SA->_coef_c;
      double c11 = a11*a11, c22 = a22*a22, c21= a21*a21;
      double b=-a11-a22,c=-c21+a11*a22;
      double   delta = b*b - 4 * c ;
      double n2=(c11+c22+c21);
      if ( n2 < 1.e-30) {
         lambda1 = lambda2=0.;
         vec1[0] = 1; vec1[1] = 0; // v.x=1,v.y=0;
        }
      else if (delta < 1.e-5*n2) // eps=1e-5
        {
          lambda1 = lambda2 = -b/2;
          vec1[0] = 1; vec1[1] = 0; // v.x=1,v.y=0;
        }
      else
        {  //    ---  construction  de 2 vecteur dans (Im ( A - D(i) Id) ortogonal
           // construction of 2 vectors in (Im ( A - D(i) Id) ortogonal
          delta = sqrt(delta);
          lambda1 = (-b-delta)/2.0;
          lambda2 = (-b+delta)/2.0;
          double v0 = a11-lambda1, v1 = a21, v2 = a22 - lambda1;
          double s0 = v0*v0 + v1*v1, s1 = v1*v1 +v2*v2;

          if (s1 < s0) {
            s0 = sqrt(s0);
            vec1[0]=v1/s0; vec1[1]=-v0/s0;  //v.x=v1/s0,v.y=-v0/s0;
          }
          else
          {
            s1=sqrt(s1);
            vec1[0]=v2/s1; vec1[1]=-v1/s1; //v.x=v2/s1,v.y=-v1/s1;
          }
      }
      // The eigenvector of lambda2.
      vec2[0] =  vec1[1];
      vec2[1] = -vec1[0];
    }
    */
    double r1 = sqrt(1./lambda1);
    double r2 = sqrt(1./lambda2);    
    //====================================================

    //return; /////////////////////////////

    double ctx, cty, radius;
    int segnum = 48; // 72

    ctx = 0.5 * (_geom->io_xmin + _geom->io_xmax);
    cty = 0.5 * (_geom->io_ymin + _geom->io_ymax);
    radius = 0.5 * sqrt((_geom->io_xmax-_geom->io_xmin)*(_geom->io_xmax-_geom->io_xmin) +
                        (_geom->io_ymax-_geom->io_ymin)*(_geom->io_ymax-_geom->io_ymin));

    // Generate the points on circle.
    double ang_start = 0.; // ang1 / 180. * detri2::PI;
    double dfi = 360. / segnum; // fabs(ang2 - ang1) / segnum;
    dfi = dfi / 180.0 * detri2::PI;

    /*
    int np = segnum + 1;
    double *ptlist = new double[np * 2];
    for (int i = 0; i < np; i++) {
      double *pp = &(ptlist[i*2]);
      pp[0] = radius * cos(ang_start + dfi * (double) i) + ctx;
      pp[1] = radius * sin(ang_start + dfi * (double) i) + cty;
    }
    */

    int firstindex = 1; // Set it to be 0 or 1.
    _disc->io_firstindex = firstindex;

    // Read the input points and weights.
    int i;
    double x, y, x1, y1, x2,y2,z2;
    int nv = segnum; //_geom->tr_steiners->objects;
    _disc->ct_in_vrts = nv;
    _disc->in_vrts = new Vertex[nv];
    for (i = 0; i < nv; i++) {
      Vertex *vrt = &(_disc->in_vrts[i]);
      vrt->init();
      //x = vrt->crd[0] = radius * cos(ang_start + dfi * (double) i) + ctx;
      //y = vrt->crd[1] = radius * sin(ang_start + dfi * (double) i) + cty;
      //vrt->crd[2] = x*x+y*y;
      // (x,y) is on the circle.
      x = radius * cos(ang_start + dfi * (double) i);
      y = radius * sin(ang_start + dfi * (double) i);
      // scale (x,y) using eigenvalues.
      x1 = x * r1;
      y1 = y * r2;
      // rotate (x,y) by the matrix [v1, v2].
      x2 = vec1[0] * x1 + vec1[1] * y1;
      y2 = vec2[0] * x1 + vec2[1] * y1;
      vrt->crd[0] = x2 + ctx;
      vrt->crd[1] = y2 + cty;
      // [2019-08-19] Here we still use Euclidean metric for generating the mesh.
      vrt->crd[2] = vrt->crd[0] * vrt->crd[0] + vrt->crd[1] * vrt->crd[1];
      vrt->idx = i + firstindex;
    }
    _disc->ct_unused_vrts=nv;
    _disc->io_xmin = _min[0];
    _disc->io_xmax = _max[1];
    _disc->io_ymin = _min[1];
    _disc->io_ymax = _max[1];
    _disc->io_diagonal = 2. * radius;

    _disc->incremental_delaunay();

    // debug only
    strcpy(_disc->io_outfilename, "/Users/si/tmp/disc_mesh");
    //_disc->save_triangulation();

    /*
    int *elist = new int[segnum * 2];
    for (int i = 0; i < segnum; i++) {
      elist[i*2] = i;
      elist[i*2+1] = i+1;
    }
    */

    _disc->tr_segs = new arraypool(sizeof(Triang), 8);
    for (i = 0; i < segnum-1; i++) {
      Triang *seg =  (Triang *) _disc->tr_segs->alloc();
      seg->init();
      seg->vrt[0] = &(_disc->in_vrts[i]);
      seg->vrt[1] = &(_disc->in_vrts[i+1]);
      seg->tag = -1;
      //seg->val = gseg->val;
      //printf("  get segment %d,%d, tag=%d, val=%g\n", seg->vrt[0]->idx, seg->vrt[1]->idx, seg->tag, seg->val);
    }
    // The last segment.
    Triang *seg =  (Triang *) _disc->tr_segs->alloc();
    seg->init();
    seg->vrt[0] = &(_disc->in_vrts[i]);
    seg->vrt[1] = &(_disc->in_vrts[0]);
    seg->tag = -1;

    _disc->recover_segments();

    // Refine mesh with min edge length.
    _disc->op_target_length = _disc->get_distance(seg->vrt[0], seg->vrt[1]);
    _disc->op_cosminangle = cos(30.0 / 180.0 * PI);
    _disc->delaunay_refinement();

    // debug only
    //strcpy(_disc->io_outfilename, "/Users/si/tmp/disc_mesh");
    //_disc->save_triangulation();
}

//=====================================================================================
/*
void OGLWidget::prepare_view()
{
    //glViewport(0,0,_w,_h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    if (_ortho) {
      //glOrtho(-_w/2, _w/2, -_h/2, _h/2, -100, 100); // Wrong, see nothing (too small)
      double dx=(_max[0]-_min[0])*1.2;
      double dy=(_max[1]-_min[1])*1.2;
      double dz=(_max[2]-_min[2])*1.2;
      double dmax = dx > dy ? dx : dy;
      //glOrtho(-dx/2., dx/2., -dy/2., dy/2., -100, 100); // stretched images
      double aspect=(float)_w/_h;
      //glOrtho(-aspect*dx/2., aspect*dx/2., -dy/2., dy/2., -dz*_zoom_max, dz*_zoom_max); // this results bad ratio
      glOrtho(-aspect*dmax/2., aspect*dmax/2., -dmax/2., dmax/2., -dz*_zoom_max, dz*_zoom_max);
    } else {
      // //gluPerspective(45, (float)_w/_h, 0.01, 100.0);
      // // Use equivalent function glFrstum()
      // // see (page 7) OpenGL Matrix Class.pdf
      //
      // double fovY = 45.0; // 30.0;
      // double aspect = (float)_w/_h;
      // double near = 0.01;
      // double far = 100.0;
      //
      // double ang = fovY/180.0*3.1415926;
      // double tagent = tan(ang/2.);
      // double hei = near * tagent;
      // double wid = hei * aspect;
      //
      // glFrustum(-wid, wid, -hei, hei, near, far);

      double dz=(_max[2]-_min[2])*1.2;
      double Near = 2.0 * dz;
      double Far = Near + 4. * dz; // 4 * dz
      gluPerspective(45, (float)_w/_h, Near, Far);
    }

    for (int i = 0; i < 16; i++) _Mp[i] = 0.;
    glGetFloatv(GL_PROJECTION_MATRIX, _Mp);

    char sbuff[256];
    if (0) {
      qDebug()<<"Projection Matrix:";
      sprintf(sbuff, " | %g  %g  %g  %g |", _Mp[0], _Mp[4], _Mp[ 8], _Mp[12]); qDebug()<<sbuff;
      sprintf(sbuff, " | %g  %g  %g  %g |", _Mp[1], _Mp[5], _Mp[ 9], _Mp[13]); qDebug()<<sbuff;
      sprintf(sbuff, " | %g  %g  %g  %g |", _Mp[2], _Mp[6], _Mp[10], _Mp[14]); qDebug()<<sbuff;
      sprintf(sbuff, " | %g  %g  %g  %g |", _Mp[3], _Mp[7], _Mp[11], _Mp[15]); qDebug()<<sbuff;
      qDebug()<<"....";
    }

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    if (!_ortho) {
      // https://www.opengl.org/archives/resources/faq/technical/viewing.htm
      //  GLU library provides the gluLookAt() function, which takes an eye position,
      //    a position to look at, and an up vector, all in object space coordinates.
      //    This function computes the inverse camera transform according to its parameters
      //    and multiplies it onto the current matrix stack.
      // see also https://www.cosc.brocku.ca/Offerings/3P98/course/lectures/3d_perspective/
      double eyeX, eyeY, eyeZ;
      double centerX, centerY, centerZ;
      double upX, upY, upZ;

      double dz=(_max[2]-_min[2])*1.2;

      eyeX = 0;
      eyeY = 0;
      eyeZ = 5. * dz; // 5.; = zoom + Near
      centerX = 0.;
      centerY = 0.;
      centerZ = 0.;
      upX = 0.;
      upY = 1.;
      upZ = 0.;

      gluLookAt(eyeX, eyeY, eyeZ,
                centerX, centerY, centerZ,
                upX, upY, upZ);

      // for (int i = 0; i < 16; i++) M[i] = 0.;
      // glGetFloatv(GL_MODELVIEW_MATRIX, M);
      // qDebug()<<"After gluLookAt Model View Matrix:";
      // sprintf(sbuff, " | %g  %g  %g  %g |", M[0], M[4], M[ 8], M[12]); qDebug()<<sbuff;
      // sprintf(sbuff, " | %g  %g  %g  %g |", M[1], M[5], M[ 9], M[13]); qDebug()<<sbuff;
      // sprintf(sbuff, " | %g  %g  %g  %g |", M[2], M[6], M[10], M[14]); qDebug()<<sbuff;
      // sprintf(sbuff, " | %g  %g  %g  %g |", M[3], M[7], M[11], M[15]); qDebug()<<sbuff;
    }

    glTranslatef(_tranX, _tranY, 0);
    glRotatef(_rotX, 1.0f, 0.0f, 0.0f);
    glRotatef(_rotY, 0.0f, 1.0f, 0.0f);
    glScalef(_zoom, _zoom, _zoom);

    for (int i = 0; i < 16; i++) _Mv[i] = 0.;
    glGetFloatv(GL_MODELVIEW_MATRIX, _Mv);

    if (0) {
      qDebug()<<"Model View Matrix:";
      sprintf(sbuff, " | %g  %g  %g  %g |", _Mv[0], _Mv[4], _Mv[ 8], _Mv[12]); qDebug()<<sbuff;
      sprintf(sbuff, " | %g  %g  %g  %g |", _Mv[1], _Mv[5], _Mv[ 9], _Mv[13]); qDebug()<<sbuff;
      sprintf(sbuff, " | %g  %g  %g  %g |", _Mv[2], _Mv[6], _Mv[10], _Mv[14]); qDebug()<<sbuff;
      sprintf(sbuff, " | %g  %g  %g  %g |", _Mv[3], _Mv[7], _Mv[11], _Mv[15]); qDebug()<<sbuff;
      qDebug()<<"....";
    }
}
*/

/*
// preparing view transformation (code from gltools)
void OGLWidget::prepare_view_transform()
{
      double xzoom = 1.2;
      double vpasp = (double)_w/(double)_h;
      double rotx,roty,rotz;
      double tranx,trany,tranz;
      double dummy,zoom;
      rotx=roty=rotz=0;
      tranx=trany=tranz=0;

      rotz = _rotY;
      rotx = _rotX;
      tranx = _tranX;
      trany = -_tranY;
      zoom = _zoom;

      double eyex,eyey,eyez;
      double centerx,centery,centerz;
      double upx,upy,upz;

      eyex=0.0;
      eyey=0.0;
      eyez=5.0;

      centerx=0.0;
      centery=0.0;
      centerz=0.0;

      upx=0.0;
      upy=1.0;
      upz=0.0;

      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();

      if (_ortho) {
        double zfac=1.2*zoom*xzoom;
        glOrtho(-vpasp*zfac, vpasp*zfac, -zfac, zfac,-zfac, zfac);
        _zNear = -zfac; _zFar = zfac;
      } else {
        gluPerspective (33*zoom*xzoom, vpasp, 4.0, 6.0);
        _zNear = 4.0; _zFar = 6.0;
      }

      // https://stackoverflow.com/questions/41935764/opengl-depth-test-is-not-working
      // Depth buffer precision is affected by the values specified for zNear and zFar.
      // The greater the ratio of zFar to zNear is, the less effective the depth buffer
      // will be at distinguishing between surfaces that are near each other.
      // If r = zFar / zNear roughtly log2(r) bits of depth buffer precision are lost.
      // Because r approaches infinity as zNear approaches 0, zNear must never be set to 0.
      glEnable(GL_DEPTH_TEST);

      // for function unproject().
      // https://stackoverflow.com/questions/3089271/converting-mouse-position-to-world-position-opengl
      glEnable(GL_DEPTH);

      for (int i = 0; i < 16; i++) _Mp[i] = 0.;
      glGetFloatv(GL_PROJECTION_MATRIX, _Mp);

      char sbuff[256];
      if (0) {
        qDebug()<<"Projection Matrix:";
        sprintf(sbuff, " | %g  %g  %g  %g |", _Mp[0], _Mp[4], _Mp[ 8], _Mp[12]); qDebug()<<sbuff;
        sprintf(sbuff, " | %g  %g  %g  %g |", _Mp[1], _Mp[5], _Mp[ 9], _Mp[13]); qDebug()<<sbuff;
        sprintf(sbuff, " | %g  %g  %g  %g |", _Mp[2], _Mp[6], _Mp[10], _Mp[14]); qDebug()<<sbuff;
        sprintf(sbuff, " | %g  %g  %g  %g |", _Mp[3], _Mp[7], _Mp[11], _Mp[15]); qDebug()<<sbuff;
        qDebug()<<"....";
      }

      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();

      if (!_ortho) {
          gluLookAt(eyex,eyey,eyez,
                centerz,centery,centerz,
                upx,upy,upz);
      }

      glTranslated(tranx,-trany,tranz);
      glRotated(rotx,1.0,0.0,0.0);
      glRotated(roty,0.0,1.0,0.0);
      glRotated(rotz,0.0,0.0,1.0);

      // Scale the objects coordinates between [-1,-1,-1], [1,1,1].
      double xsize= (_max[0]-_min[0]);
      double ysize= (_max[1]-_min[1]);
      double zsize= (_max[2]-_min[2]);
      double aspz=1.0;
      double aspxy=1.0;

      double scalefac=xsize;
      if (ysize>scalefac) scalefac=ysize;
      if (zsize>scalefac) scalefac=zsize;

      glScalef(1.0/(scalefac*aspxy),aspxy/scalefac,aspz/scalefac);

      // shift volume center into origin
      // [Comment]: so we see the objects in the middle.
      glTranslatef(
                 -0.5*(_max[0]+_min[0]),
                 -0.5*(_max[1]+_min[1]),
                 -0.5*(_max[2]+_min[2])
                 );

      for (int i = 0; i < 16; i++) _Mv[i] = 0.;
      glGetFloatv(GL_MODELVIEW_MATRIX, _Mv);

      if (0) {
        qDebug()<<"Model View Matrix:";
        sprintf(sbuff, " | %g  %g  %g  %g |", _Mv[0], _Mv[4], _Mv[ 8], _Mv[12]); qDebug()<<sbuff;
        sprintf(sbuff, " | %g  %g  %g  %g |", _Mv[1], _Mv[5], _Mv[ 9], _Mv[13]); qDebug()<<sbuff;
        sprintf(sbuff, " | %g  %g  %g  %g |", _Mv[2], _Mv[6], _Mv[10], _Mv[14]); qDebug()<<sbuff;
        sprintf(sbuff, " | %g  %g  %g  %g |", _Mv[3], _Mv[7], _Mv[11], _Mv[15]); qDebug()<<sbuff;
        qDebug()<<"....";
      }
}
*/

// /Users/si/tetview2_wrk/pdelib_wrk/modules/graphic/gltools/glviewer.cxx
// In function _wdraw(w, h)
void OGLWidget::prepare_view_transform()
{
    double vpasp = (double)_w/(double)_h;
    double rotx,roty,rotz;
    double tranx,trany,tranz;
    double dummy,zoom;
    rotx=roty=rotz=0;
    tranx=trany=tranz=0;

    //if (_guidim>2) mouse_control("rot")->get_state(&rotz,&rotx);
    //mouse_control("tran")->get_state(&tranx,&trany);
    //mouse_control("zoom")->get_state(&dummy,&zoom);

    roty = _rotZ;
    rotz = _rotY;
    rotx = _rotX;    
    tranx = _tranX;
    trany = -_tranY;
    zoom = _zoom;

    double eyex,eyey,eyez;
    double centery,centerz;
    double upx,upy,upz;
    double Near,Far;

    Near = 4.0*zoom;
    Far = Near + zoom*2;

    eyex=0.0;
    eyey=0.0;
    eyez= zoom + Near;
    centery=0.0;
    centerz=0.0;

    upx=0.0;
    upy=1.0;
    upz=0.0;

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    if (_ortho) {
      glOrtho(-1*vpasp, 1*vpasp, -1, 1, -zoom, zoom);
      _zNear = -zoom; _zFar = zoom;
    } else {
      glFrustum(-1*vpasp, 1*vpasp, -1, 1, Near, Far);
      _zNear = Near; _zFar = Far;
    }

    // https://stackoverflow.com/questions/41935764/opengl-depth-test-is-not-working
    // Depth buffer precision is affected by the values specified for zNear and zFar.
    // The greater the ratio of zFar to zNear is, the less effective the depth buffer
    // will be at distinguishing between surfaces that are near each other.
    // If r = zFar / zNear roughtly log2(r) bits of depth buffer precision are lost.
    // Because r approaches infinity as zNear approaches 0, zNear must never be set to 0.
    glEnable(GL_DEPTH_TEST);

    // for function unproject().
    // https://stackoverflow.com/questions/3089271/converting-mouse-position-to-world-position-opengl
    glEnable(GL_DEPTH);

    for (int i = 0; i < 16; i++) _Mp[i] = 0.;
    glGetFloatv(GL_PROJECTION_MATRIX, _Mp);

    char sbuff[256];
    if (0) {
      qDebug()<<"Projection Matrix:";
      sprintf(sbuff, " | %g  %g  %g  %g |", _Mp[0], _Mp[4], _Mp[ 8], _Mp[12]); qDebug()<<sbuff;
      sprintf(sbuff, " | %g  %g  %g  %g |", _Mp[1], _Mp[5], _Mp[ 9], _Mp[13]); qDebug()<<sbuff;
      sprintf(sbuff, " | %g  %g  %g  %g |", _Mp[2], _Mp[6], _Mp[10], _Mp[14]); qDebug()<<sbuff;
      sprintf(sbuff, " | %g  %g  %g  %g |", _Mp[3], _Mp[7], _Mp[11], _Mp[15]); qDebug()<<sbuff;
      qDebug()<<"....";
    }

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    if (!_ortho) {
        gluLookAt(eyex,eyey,eyez,
              centerz,centery,centerz,
              upx,upy,upz);
    }

    //glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
    //int outline=submenu_entry_state("View","Outline");
    glScaled(zoom, zoom, zoom);

    //if (_w>_h)
    //   glTranslated(-((double)_w/(double)_h-1.0),0,0);

    glTranslated(tranx,-trany,tranz);
    glRotated(rotx,1.0,0.0,0.0);
    glRotated(roty,0.0,1.0,0.0);
    glRotated(rotz,0.0,0.0,1.0);

    // Scale the objects coordinates between [-1,-1,-1], [1,1,1].
    double xsize=/*scale[0]*/ (_max[0]-_min[0]);
    double ysize=/*scale[1]*/ (_max[1]-_min[1]);
    double zsize=/*scale[2]*/ (_max[2]-_min[2]);
    double aspz=1.0;
    double aspxy=1.0;

    double scalefac=xsize;
    if (ysize>scalefac) scalefac=ysize;
    if (zsize>scalefac) scalefac=zsize;

    glScalef(1.0/(scalefac*aspxy),aspxy/scalefac,aspz/scalefac);

    // shift volume center into origin
    // [Comment]: so we see the objects in the middle.
    glTranslatef(
               -0.5*(_max[0]+_min[0]),
               -0.5*(_max[1]+_min[1]),
               -0.5*(_max[2]+_min[2])
               );

    for (int i = 0; i < 16; i++) _Mv[i] = 0.;
    glGetFloatv(GL_MODELVIEW_MATRIX, _Mv);

    if (0) {
      qDebug()<<"Model View Matrix:";
      sprintf(sbuff, " | %g  %g  %g  %g |", _Mv[0], _Mv[4], _Mv[ 8], _Mv[12]); qDebug()<<sbuff;
      sprintf(sbuff, " | %g  %g  %g  %g |", _Mv[1], _Mv[5], _Mv[ 9], _Mv[13]); qDebug()<<sbuff;
      sprintf(sbuff, " | %g  %g  %g  %g |", _Mv[2], _Mv[6], _Mv[10], _Mv[14]); qDebug()<<sbuff;
      sprintf(sbuff, " | %g  %g  %g  %g |", _Mv[3], _Mv[7], _Mv[11], _Mv[15]); qDebug()<<sbuff;
      qDebug()<<"....";
    }
}

//============================================================

void OGLWidget::draw_objects()
{
    //qDebug() << "draw_objects, _object_valid =" << _object_valid << endl;

    if (_list_base==0) // initialize display lists & create them as empty ones
    {
      _list_base=glGenLists(11); // GLV_MAX_LISTS
      //qDebug() << "init display list, _list_base=" <<_list_base<< endl;
      for(int i=0;i<11;i++) // GLV_MAX_LISTS
      {
        glNewList(_list_base+i,GL_COMPILE);
        glEndList();
        _lists[i]=i;
        _visible[i]=1;
      }
    }

    //==================================================================
    /*
    // preparing view (code from gltools)
    {
      double xzoom = 1.2;
      double vpasp = (double)_w/(double)_h;
      double rotx,roty,rotz;
      double tranx,trany,tranz;
      double dummy,zoom;
      rotx=roty=rotz=0;
      tranx=trany=tranz=0;

      rotz = _rotY;
      rotx = _rotX;
      tranx = _tranX;
      trany = -_tranY;
      zoom = _zoom;

      double eyex,eyey,eyez;
      double centerx,centery,centerz;
      double upx,upy,upz;

      eyex=0.0;
      eyey=0.0;
      eyez=5.0;

      centerx=0.0;
      centery=0.0;
      centerz=0.0;

      upx=0.0;
      upy=1.0;
      upz=0.0;

      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();

      if (_ortho) {
        double zfac=1.5*zoom*xzoom;
        glOrtho(-vpasp*zfac, vpasp*zfac, -zfac, zfac,-zfac,zfac);
      } else {
        gluPerspective (33*zoom*xzoom, vpasp, 4.0, 6.0);
      }

      for (int i = 0; i < 16; i++) _Mp[i] = 0.;
      glGetFloatv(GL_PROJECTION_MATRIX, _Mp);

      char sbuff[256];
      if (0) {
        qDebug()<<"Projection Matrix:";
        sprintf(sbuff, " | %g  %g  %g  %g |", _Mp[0], _Mp[4], _Mp[ 8], _Mp[12]); qDebug()<<sbuff;
        sprintf(sbuff, " | %g  %g  %g  %g |", _Mp[1], _Mp[5], _Mp[ 9], _Mp[13]); qDebug()<<sbuff;
        sprintf(sbuff, " | %g  %g  %g  %g |", _Mp[2], _Mp[6], _Mp[10], _Mp[14]); qDebug()<<sbuff;
        sprintf(sbuff, " | %g  %g  %g  %g |", _Mp[3], _Mp[7], _Mp[11], _Mp[15]); qDebug()<<sbuff;
        qDebug()<<"....";
      }

      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();

      if (!_ortho) {
          gluLookAt(eyex,eyey,eyez,
                centerz,centery,centerz,
                upx,upy,upz);
      }

      glTranslated(tranx,-trany,tranz);
      glRotated(rotx,1.0,0.0,0.0);
      glRotated(roty,0.0,1.0,0.0);
      glRotated(rotz,0.0,0.0,1.0);

      // Scale the objects coordinates between [-1,-1,-1], [1,1,1].
      double xsize= (_max[0]-_min[0]);
      double ysize= (_max[1]-_min[1]);
      double zsize= (_max[2]-_min[2]);
      double aspz=1.0;
      double aspxy=1.0;

      double scalefac=xsize;
      if (ysize>scalefac) scalefac=ysize;
      if (zsize>scalefac) scalefac=zsize;

      glScalef(1.0/(scalefac*aspxy),aspxy/scalefac,aspz/scalefac);

      // shift volume center into origin
      // [Comment]: so we see the objects in the middle.
      glTranslatef(
                 -0.5*(_max[0]+_min[0]),
                 -0.5*(_max[1]+_min[1]),
                 -0.5*(_max[2]+_min[2])
                 );

      for (int i = 0; i < 16; i++) _Mv[i] = 0.;
      glGetFloatv(GL_MODELVIEW_MATRIX, _Mv);

      if (0) {
        qDebug()<<"Model View Matrix:";
        sprintf(sbuff, " | %g  %g  %g  %g |", _Mv[0], _Mv[4], _Mv[ 8], _Mv[12]); qDebug()<<sbuff;
        sprintf(sbuff, " | %g  %g  %g  %g |", _Mv[1], _Mv[5], _Mv[ 9], _Mv[13]); qDebug()<<sbuff;
        sprintf(sbuff, " | %g  %g  %g  %g |", _Mv[2], _Mv[6], _Mv[10], _Mv[14]); qDebug()<<sbuff;
        sprintf(sbuff, " | %g  %g  %g  %g |", _Mv[3], _Mv[7], _Mv[11], _Mv[15]); qDebug()<<sbuff;
        qDebug()<<"....";
      }
    }
    */
    //==================================================================

    //if (!(_mouse_down&& hide_on_move))
    //if (!_mouse_down)
    {
      if (!_object_valid)
      {
        //qDebug() << "compiling and executing display list 0" << endl;
        glNewList(_list_base+_lists[0],GL_COMPILE_AND_EXECUTE);
        if (_draw_object) //if (_draw_object && _object_data)
          (this->*_draw_object)(); // (this,_object_data);
        glEndList();
        _object_valid=1;
      }
      else
      {
        //qDebug() << "calling displaylist 0" << endl;
        if (_draw_object)
          glCallList(_list_base+_lists[0]);
      }
    }

      //for(int i=1;i<GLV_MAX_LISTS;i++)
      //  if (_visible[i])
      //    glCallList(_list_base+_lists[i]);
}

void OGLWidget::draw_bbox()
{
    //printf("debug:  draw_bbox() is called.\n");

    if (showBbox->isChecked()) {
      glLineWidth(2);
      glColor4f(1.0, 0.0, 0.0, 0.3);

      glBegin(GL_LINE_LOOP);
        glVertex3f(_min[0],_min[1],_min[2]);
        glVertex3f(_max[0],_min[1],_min[2]);
        glVertex3f(_max[0],_max[1],_min[2]);
        glVertex3f(_min[0],_max[1],_min[2]);
      glEnd();

      glBegin(GL_LINE_LOOP);
        glVertex3f(_min[0],_min[1],_max[2]);
        glVertex3f(_max[0],_min[1],_max[2]);
        glVertex3f(_max[0],_max[1],_max[2]);
        glVertex3f(_min[0],_max[1],_max[2]);
      glEnd();

      glBegin(GL_LINES);
        glVertex3f(_min[0],_min[1],_min[2]);
        glVertex3f(_min[0],_min[1],_max[2]);

        glVertex3f(_max[0],_min[1],_min[2]);
        glVertex3f(_max[0],_min[1],_max[2]);

        glVertex3f(_max[0],_max[1],_min[2]);
        glVertex3f(_max[0],_max[1],_max[2]);

        glVertex3f(_min[0],_max[1],_min[2]);
        glVertex3f(_min[0],_max[1],_max[2]);
      glEnd();
    }

    if (showPlane->isChecked()) {
      // Draw the cut-plane
      //glvMouseControl *mc = mouse_control("plane");
      //if (mc != NULL) {
        double eps=5.0e-2*(_max[2]-_min[2]);
        //double dummy, plane_d = 0.0;
        //int plane_dir = 2; // default is dir Z.
        //if (dirZAct->isChecked()) plane_dir = 2; // z-section
        //if (dirYAct->isChecked()) plane_dir = 1; // y-section
        //if (plane_dir == 0) { // X direction
        //  mc->get_state(&plane_d, &dummy); // use x value
        //} else {
        //  mc->get_state(&dummy, &plane_d); // use y value
        //}
        //double plane_d=_min[_plane_dir]+_plane_d*(_max[_plane_dir]-_min[_plane_dir]);
        double plane_d=0; //_min[_plane_dir]+_plane_d*(_max[_plane_dir]-_min[_plane_dir]);
        //plane_d = -plane_d;
        glEnable(GL_BLEND);
        glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
        glColor4d(0.5,0.5,0.5,0.45); // a = 1.0 (no transparency)
        glBegin(GL_QUADS);
        if (_plane_dir == 2) { // z-section
          glNormal3f(0,0,1);
          glVertex3f(_min[0]-eps,_min[1]-eps,plane_d+1.0e-2*eps);
          glVertex3f(_max[0]+eps,_min[1]-eps,plane_d+1.0e-2*eps);
          glVertex3f(_max[0]+eps,_max[1]+eps,plane_d+1.0e-2*eps);
          glVertex3f(_min[0]-eps,_max[1]+eps,plane_d+1.0e-2*eps);
        } else if (_plane_dir == 1) { // y-section
          glNormal3f(0,1,0);
          glVertex3f(_min[0]-eps,plane_d+1.0e-2*eps,_min[2]-eps);
          glVertex3f(_max[0]+eps,plane_d+1.0e-2*eps,_min[2]-eps);
          glVertex3f(_max[0]+eps,plane_d+1.0e-2*eps,_max[2]+eps);
          glVertex3f(_min[0]-eps,plane_d+1.0e-2*eps,_max[2]+eps);
        } else if (_plane_dir == 0) { // z-section
          glNormal3f(1,0,0);
          glVertex3f(plane_d+1.0e-2*eps,_min[1]-eps,_min[2]-eps);
          glVertex3f(plane_d+1.0e-2*eps,_max[1]+eps,_min[2]-eps);
          glVertex3f(plane_d+1.0e-2*eps,_max[1]+eps,_max[2]+eps);
          glVertex3f(plane_d+1.0e-2*eps,_min[1]-eps,_max[2]+eps);
        }
        glEnd();
        glDisable(GL_BLEND);
      //} // if (mc != NULL)
    } // _show_cut_plane

    if (showBbox->isChecked()) {
      // Show bbox coordinates.
      double l = _min[0]; // left
      double r = _max[0]; // right
      double b = _min[1]; // bottom
      double t = _max[1]; // top
      double n = _min[2]; // near
      double f = _max[2]; // far

      // object coordiates
      double P[4] = {l, b, n, 1.};
      double Q[4] = {r, b, n, 1.};
      double R[4] = {l, t, n, 1.};
      double S[4] = {r, t, f, 1.};

      QPainter painter(this);
      //setFont(QFont("Times", 24));
      //QColor color(255,223,127);
      //glColor4f(color.redF(), color.greenF(), color.blueF(), color.alphaF());
      painter.setPen(Qt::gray);
      painter.setFont(QFont("Arial", 16));

      char str[100];
      sprintf(str, "%.2g,%.2g,%.2g", P[0], P[1], P[2]);
      renderText(&painter, P[0], P[1], P[2], QString(str));
      sprintf(str, "%.2g,%.2g,%.2g", Q[0], Q[1], Q[2]);
      renderText(&painter, Q[0], Q[1], Q[2], QString(str));
      sprintf(str, "%.2g,%.2g,%.2g", R[0], R[1], R[2]);
      renderText(&painter, R[0], R[1], R[2], QString(str));
      sprintf(str, "%.2g,%.2g,%.2g", S[0], S[1], S[2]);
      renderText(&painter, S[0], S[1], S[2], QString(str));

      painter.end();
    }

    /*
    {
        QPainter  painter(this);
        painter.setPen(Qt::gray);
        painter.setFont(QFont("Arial", 24));
        painter.drawText(0, 0, width(), height(), Qt::AlignCenter, "Hello World!");
        painter.end();
    }
    */
}

void OGLWidget::draw_axes()
{
    if (!showAxis->isChecked()) return;

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    glOrtho(-_w/2, _w/2, -_h/2, _h/2, -100, 100);

    for (int i = 0; i < 16; i++) _Mp[i] = 0.;
    glGetFloatv(GL_PROJECTION_MATRIX, _Mp);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // Draw x-, y-, z-axis.
    glTranslatef(-_w/2 + 40, -_h/2 + 40, 0.0f);
    //mouse_control("tran")->get_state(&tranx,&trany);

    // [2019-08-09] Tried
    //glRotatef(_rotX, 1.0f, 0.0f, 0.0f);
    //glRotatef(_rotY, 0.0f, 1.0f, 0.0f);
    double rotx,roty,rotz;
    //if (_guidim>2) mouse_control("rot")->get_state(&rotz,&rotx);
    roty = _rotZ;
    rotz = _rotY;
    rotx = _rotX;
    glRotated(rotx,1.0,0.0,0.0);
    glRotated(roty,0.0,1.0,0.0);
    glRotated(rotz,0.0,0.0,1.0);

    for (int i = 0; i < 16; i++) _Mv[i] = 0.;
    glGetFloatv(GL_MODELVIEW_MATRIX, _Mv);

    glLineWidth(3);
    glBegin(GL_LINES);
        glColor4f(1.0, 0.0, 0.0, 0.3);
        glVertex3f(0, 0, 0);
        glVertex3f(30, 0, 0);

        glColor4f(0.0, 1.0, 0.0, 0.3);
        glVertex3f(0, 0, 0);
        glVertex3f(0, 30, 0);

        glColor4f(0.0, 0.0, 1.0, 0.3);
        glVertex3f(0, 0, 0);
        glVertex3f(0, 0, 30);
    glEnd();

    QPainter painter(this);
    setFont(QFont("Times", 12));
    QColor color(255,223,127);
    glColor4f(color.redF(), color.greenF(), color.blueF(), color.alphaF());
    painter.setPen(color); // Qt::gray
    //painter.setFont(QFont("Arial", 16));

    renderText(&painter, 30., 0., 0., QString("X"));
    renderText(&painter, 0., 30., 0., QString("Y"));
    renderText(&painter, 0., 0., 30., QString("Z"));

    //painter.setPen(color); // Qt::gray
    painter.setFont(QFont("Arial", 16));
    if (_ortho) {
      painter.drawText(30, 20, QString("Orthogonal"));
    } else {
      painter.drawText(30, 20, QString("Perspective"));
    }

    painter.end();
}

//==============================================================================

void OGLWidget::initializeGL()
{
    /*
    glClearColor(0,0,0,1);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHT0);
    //glEnable(GL_LIGHTING);
    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);
    */

    printf("debug:  OGLWidget::initializeGL() is called.\n");

    //glShadeModel( GL_SMOOTH );
    double r, g, b, a;
    _bgcolor.getRgbF(&r, &g, &b, &a);
    glClearColor(r, g, b, a); //RGBA
    glClearDepth(1.0f);

    // https://stackoverflow.com/questions/41935764/opengl-depth-test-is-not-working
    //glEnable(GL_DEPTH_TEST);  // no use to call it too early
    glDepthFunc(GL_LEQUAL);

    //glEnable(GL_BLEND);
    //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glBlendFunc(GL_ONE_MINUS_SRC_ALPHA, GL_SRC_ALPHA);

    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    //glEnable(GL_LIGHT0);
    //glEnable(GL_LIGHTING);

    //glEnable(GL_LINE_SMOOTH);
    //glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
}

void OGLWidget::resizeGL(int w, int h)
{
    _w = w;
    _h = h;
    glViewport(0,0,w,h);
}

void OGLWidget::paintGL()
{
    //qDebug() << "paintGL is called";

    //glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    //glLoadIdentity();

    draw_axes();
    prepare_view_transform();
    draw_objects();
    //draw_labels();
    draw_selection();    
    draw_bbox();
}
