#ifndef OGLWIDGET_H
#define OGLWIDGET_H

#include <QMainWindow>
#include <QWidget>
#include <QOpenGLWidget>
#include <QPainter>
#include <QMenuBar>
#include <QToolBar>
#include <QStatusBar>
#include <QLabel>
#include <QDebug>
#include <math.h>

#include "detri2/detri2.h"

//class OGLWidget;
// /** \brief Callback for drawing object and legend */
//typedef void (*glvDrawCallback)(OGLWidget *viewer, void* userdata);
//typedef void (OGLWidget::*DrawCallback)();

class OGLWidget : public QOpenGLWidget
{
    Q_OBJECT
public:
    OGLWidget(void *scribbleArea,
              QMenuBar* menubar,
              QStatusBar* statusbar,
              QToolBar* toolbar_top,
              QToolBar* toolbar_left,
              QWidget *parent = 0);
    ~OGLWidget();

    // These are pointers to items in the mainwindow
    QMenuBar*   _menubar;
    QStatusBar* _statusbar;
    QToolBar*   _toolbar_top;
    QToolBar*   _toolbar_left;

    QLabel*     m_statusLeft;
    QLabel*     m_statusMiddle;
    QLabel*     m_statusRight;

    void       *_scribbleArea;

    detri2::Triangulation *_disc;
    double     _zshift; // default 0.25
    double     _ccent[2], _radius;

    // Quadratic form coefficients.
    //double _a11, _a21, _a22;

    // A Callback function pointer to class member functions.
    typedef void (OGLWidget::*DrawCallback)();

    //glvDrawCallback _draw_object;
    DrawCallback    _draw_object;
    long            _lists[11]; // GLV_MAX_LISTS
    int             _visible[11];
    int             _object_valid;
    GLint           _list_base;    /**< OpenGL Display list base */

    double  _min[3], _max[3]; // object bounding box
    double  _dmax; // maximum size of x, y, z.

    bool    _ortho;
    int     _w, _h; // set in resizeGL()
    float   _Mv[16], _Mp[16]; // set in view_transformation();

    double  _transparency; // (0., 1.0), 1.0 means no transparency.
    double  _shrink;
    double  _plane_normal[3], _plane_d; // a,b,c,d
    int     _plane_dir; // 0 (x), 1 (y), 2 (z)    

    enum MouseActive {eRot, eTrans, eShrink, eCutPlane, eSelect} _mouse_act;

    // view transformations
    double  _rotX, _rotY, _rotZ;  // rotation angles, in degrees
    double  _tranX, _tranY;
    double  _zoom;
    double  _zoom_max;
    double  _zNear, _zFar;
    int     _prevX, _prevY;  // mouse position, in screen coordinates    

    // view preferences
    int     _vertex_size;
    int     _edge_width;
    int     _segment_width;
    QColor  _bgcolor;
    QColor  _vertexcolor;
    QColor  _edgecolor;    
    QColor  _facetcolor;
    QColor  _selectcolor;
    QColor  _circlecolor;
    QColor  _weightcolor;
    QColor  _polecolor;
    QColor  _polarcolor;

    // File options
    QString _file_path;

    // Selection test
    detri2::Vertex _pos, _ray;
    detri2::arraypool *_sel_vrts;
    detri2::arraypool *_sel_edges;

public:
    // file menu
    QAction*    openAct;
    QAction*    saveAct;
    QAction*    closeAct;

    // Mouse actions
    QAction*    rotateAct;
    QAction*    transAct;
    QAction*    shrinkAct;
    QAction*    planeAct;
    QAction*    zoomRectAct;
    QAction*    selectAct;
    QAction*    selectVertexAct;
    QAction*    selectEdgeAct;
    QAction*    selectFaceAct;

    // view options
    QAction*    resetVolumeAct;
    QAction*    showVertices;
    QAction*    showVertexLabels;
    QAction*    showEdges;
    QAction*    showFacets;
    QAction*    showSegments;
    QAction*    showBbox;
    QAction*    showPlane;
    QAction*    showAxis;
    QAction*    showTransparent;
    QAction*    enableCutPlane;
    QAction*    invertCutPlane;
    QAction*    enableShrink;    

    // debug options
    QAction*    queryVertices;
    QAction*    queryEdges;
    QAction*    queryTriangles;
    QAction*    queryTetra;
    QAction*    showqueryEdges;
    QAction*    clearSelectionAct;

    //=======================================
    QAction*    viewWeightsAct;
    QAction*    viewPolesAct;
    QAction*    viewPolarsAct;
    QAction*    viewCircleAct;
    QAction*    viewGeometryAct;
    QAction*    viewParaboildAct;
    QAction*    viewlifttransformAct;
    QAction*    viewprojectionAct;
    QAction*    viewVoronoiAct;
    QAction*    setZliftAct;
    //=======================================

    void    createActions();
    void    createMenus();

    void    reset_volume();
    void    reset_view_transformations();
    void    reset_view_preferences();

    void    projection(double Ox, double Oy, double Oz, int* Wx, int* Wy);
    void    unproject(int Wx, int Wy, double *Ox, double *Oy, double *Oz);
    void    raycasting(int mouse_x, int mouse_y, double *Rx, double *Ry, double *Rz);
    void    renderText(QPainter *painter, double Ox, double Oy, double Oz, const QString text);

    void    triangulate_disc();
    void    triangulate_ellipse();

public slots:
    void    ShowContextMenu(const QPoint& pos);

    void    openFile();
    void    saveFile();
    void    closeFile();
    void    open_preferDialog();

    void    mouse_act();
    void    view_update_act();
    void    set_zlift_act();
    void    reset_volume_act();
    void    update_act();
    void    query_act();
    void    clear_select_act();

    // Callbacks (plot.cpp)
    void    draw_vertices();
    void    draw_weights();
    void    draw_poles();
    void    draw_polars();
    void    draw_circle();
    void    draw_paraboloid();
    void    draw_triangulation();
    void    draw_scene();
    void    plot3d();   // tetviewmain(), set callbacks

    // paintGL()
    void    prepare_view_transform();
    void    draw_axes();
    void    draw_objects();
    //void    draw_labels();
    void    draw_selection();
    void    draw_bbox();

protected:
    virtual void    keyPressEvent(QKeyEvent *event);

    virtual void    mousePressEvent(QMouseEvent *);
    virtual void    mouseReleaseEvent(QMouseEvent *);
    virtual void    mouseMoveEvent(QMouseEvent *);
    virtual void    wheelEvent(QWheelEvent *);

    void    initializeGL();
    void    resizeGL(int w, int h);
    void    paintGL();
};

class View3dWidget : public QMainWindow
{
    Q_OBJECT
public:

    View3dWidget(void* draw_data, QWidget *parent = NULL);

    OGLWidget  *_glpanel;
};

#endif // OGLWIDGET_H
